from collections.abc import ItemsView, Iterator, KeysView, ValuesView

import pydantic


class ListOf[T](pydantic.RootModel[list[T]]):
    """
    Pydantic RootModel for representing a list of other models.
    """

    def __iter__(self) -> Iterator[T]:  # type:ignore[override]
        """Iterate over the items in the list."""
        return iter(self.root)

    def __getitem__(self, item: int) -> T:
        """Get an item by index."""
        return self.root[item]

    def __len__(self) -> int:
        """Get the length of the list."""
        return len(self.root)


class DictOf[K, V](pydantic.RootModel[dict[K, V]]):
    """
    Pydantic RootModel for representing a dict
    """

    def __getitem__(self, key: K) -> V:
        """Get a value by key."""
        return self.root[key]

    def __setitem__(self, key: K, value: V) -> None:
        """Set a value for a key."""
        self.root[key] = value

    def __delitem__(self, key: K) -> None:
        """Delete a key-value pair."""
        del self.root[key]

    def __contains__(self, key: K) -> bool:
        """Check if a key exists in the dict."""
        return key in self.root

    def __iter__(self) -> Iterator[K]:  # type:ignore[override]
        """Iterate over the keys in the dict."""
        return iter(self.root)

    def __len__(self) -> int:
        """Get the number of key-value pairs."""
        return len(self.root)

    def keys(self) -> KeysView[K]:
        """Get a view of the dict's keys."""
        return self.root.keys()

    def values(self) -> ValuesView[V]:
        """Get a view of the dict's values."""
        return self.root.values()

    def items(self) -> ItemsView[K, V]:
        """Get a view of the dict's key-value pairs."""
        return self.root.items()

    def get(self, key: K, default: V | None = None) -> V | None:
        """Get a value by key with optional default."""
        return self.root.get(key, default)

    def setdefault(self, key: K, default: V) -> V:
        """Get a value by key, setting it to default if not present."""
        return self.root.setdefault(key, default)

    def update(self, other: dict[K, V]) -> None:
        """Update the dict with key-value pairs from another dict."""
        self.root.update(other)

    def pop(self, key: K, default: V | None = None) -> V | None:
        """Remove and return a value by key with optional default."""
        return self.root.pop(key, default)

    def popitem(self) -> tuple[K, V]:
        """Remove and return an arbitrary key-value pair."""
        return self.root.popitem()

    def clear(self) -> None:
        """Remove all key-value pairs from the dict."""
        self.root.clear()



from collections.abc import Callable
from typing import Concatenate, assert_type, cast, get_args, get_origin, get_type_hints
import functools

from django.contrib.auth.models import AnonymousUser, User
from django.http import HttpRequest, HttpResponse, HttpResponseForbidden, QueryDict
import pydantic
import pydantic_core

from ..result import as_result
from .requests import (
    AnyUser,
    AuthdEmptyTypedRequest,
    AuthdTypedRequest,
    AuthdTypedRequestBody,
    AuthdTypedRequestQuery,
    EmptyTypedRequest,
    TypedRequest,
    TypedRequestBody,
    TypedRequestQuery,
)
from .responses import APIResponse
from .serializers import Empty


def _prep_querydict_for_pydantic(queryDict: QueryDict) -> dict[str, str | list[str]]:
    data: dict[str, str | list[str]] = {
        key: (val_list[0] if len(val_list) == 1 else val_list)
        for key, val_list in ((key, queryDict.getlist(key)) for key in queryDict.keys())
        if len(val_list) > 0
    }
    return data


@as_result(pydantic_core.ValidationError)
def _parse_request_querystring[_QST: pydantic.BaseModel](
    model: type[_QST],
    request: HttpRequest,
) -> _QST:
    data = _prep_querydict_for_pydantic(request.GET)
    return model.model_validate(data)


@as_result(pydantic_core.ValidationError)
def _parse_request_body[_BT: pydantic.BaseModel](
    model: type[_BT],
    request: HttpRequest,
) -> _BT:
    ct = request.headers.get("Content-Type")
    if ct is not None:
        ct = ct.split(";")[0].strip()
    if ct == "application/json":
        return model.model_validate_json(request.body)
    if ct in ("multipart/form-data", "application/x-www-form-urlencoded"):
        data = _prep_querydict_for_pydantic(request.POST)
        return model.model_validate(data)
    return model.model_validate({})


def _infer_request_types[
    **P,
    R: HttpResponse,
    _QST: pydantic.BaseModel,
    _BT: pydantic.BaseModel,
    _UT: AnyUser,
](
    fn: Callable[Concatenate[TypedRequest[_QST, _BT, _UT], P], R],
) -> tuple[
    type[_QST],  # Querystring model class
    type[_BT],  # Body model class
    bool,  # Requires authenticated user?
]:
    request_type = get_type_hints(fn).get("request") or EmptyTypedRequest
    origin = get_origin(request_type)
    if origin is None:
        origin = request_type
    args = get_args(request_type)

    # Auth-required types
    if origin is AuthdEmptyTypedRequest:
        return cast(type[_QST], Empty), cast(type[_BT], Empty), True

    if origin is AuthdTypedRequestQuery:
        return cast(type[_QST], args[0]), cast(type[_BT], Empty), True

    if origin is AuthdTypedRequestBody:
        return cast(type[_QST], Empty), cast(type[_BT], args[0]), True

    if origin is AuthdTypedRequest:
        return cast(type[_QST], args[0]), cast(type[_BT], args[1]), True

    # Auth-not-required types
    if origin is EmptyTypedRequest:
        return cast(type[_QST], Empty), cast(type[_BT], Empty), False

    if origin is TypedRequestQuery:
        return cast(type[_QST], args[0]), cast(type[_BT], Empty), False

    if origin is TypedRequestBody:
        return cast(type[_QST], Empty), cast(type[_BT], args[0]), False

    if origin is TypedRequest:
        return cast(type[_QST], args[0]), cast(type[_BT], args[1]), False

    raise TypeError(f"Failed to infer request types for {fn}")


def validate[
    **P,
    R: HttpResponse,
    _QST: pydantic.BaseModel,
    _BT: pydantic.BaseModel,
    _UT: AnyUser,
](
    query_model: type[_QST] | None = None,
    body_model: type[_BT] | None = None,
) -> Callable[
    [
        Callable[
            Concatenate[TypedRequest[_QST, _BT, _UT], P],
            R,
        ]
    ],
    Callable[
        Concatenate[HttpRequest, P],
        R | APIResponse[_QST] | APIResponse[_BT] | HttpResponseForbidden,
    ],
]:
    """
    Django view decorator which validates and transforms the incoming request
    data into a `TypedRequest` object (a subclass of HttpRequest). Infers the
    Pydantic models used for validation from the type hints on the view itself.
    For that to work correctly, the `request` param should be typed as
    `TypedRequest[Foo, Bar]` or one of the aliases to TypedRequest defined
    above.

    Args:
        query_model: Optional explicit model for querystring validation.
            Overrides annotation-inferred type when provided.
        body_model: Optional explicit model for request body validation.
            Overrides annotation-inferred type when provided.
    """

    def decorator(
        fn: Callable[Concatenate[TypedRequest[_QST, _BT, _UT], P], R],
    ) -> Callable[
        Concatenate[HttpRequest, P],
        R | APIResponse[_QST] | APIResponse[_BT] | HttpResponseForbidden,
    ]:
        # Extract validation models from annotations
        inferred_qs, inferred_body, requires_auth = _infer_request_types(fn)
        # Use explicit models if provided, otherwise use inferred
        qs_model_final = query_model if query_model is not None else inferred_qs
        body_model_final = body_model if body_model is not None else inferred_body

        @functools.wraps(fn)
        def wrapper(
            request: HttpRequest,
            *args: P.args,
            **kwargs: P.kwargs,
        ) -> R | APIResponse[_QST] | APIResponse[_BT] | HttpResponseForbidden:
            # Check authentication. This isn't intended as a full replacement
            # for more granular checks, like against specific model permissions.
            # It's meant to work in conjunction with those—the main job of this
            # code is to give the wrapped view a more specific type hint for
            # the `request.user` property.
            req_user: User | AnonymousUser = getattr(request, "user", AnonymousUser())
            if requires_auth:
                if req_user.is_anonymous:
                    return HttpResponseForbidden()
                assert_type(req_user, User)

            # Validate querystring
            qs_result = _parse_request_querystring(qs_model_final, request)
            if qs_result.is_err:
                return APIResponse(qs_result)

            # Validate body
            body_result = _parse_request_body(body_model_final, request)
            if body_result.is_err:
                return APIResponse(body_result)

            # Construct the typed request
            typed_req = TypedRequest[_QST, _BT, _UT](
                request=request,
                user=req_user,  # type:ignore[arg-type]
                qs=qs_result.ok(),
                body=body_result.ok(),
            )
            # Call the inner view with the new validated request data
            return fn(typed_req, *args, **kwargs)

        return wrapper

    return decorator



import uuid

from django.contrib.auth.models import User
from django.test import RequestFactory, TestCase
import pydantic_core

from thelabtyping.api.responses import APIResponse
from thelabtyping.api.status import Status
from thelabtyping.result import Result

from ..sampleapp.serializers import APIUser


class APIResponseTest(TestCase):
    def _get_api_user(
        self,
        valid: bool = True,
    ) -> tuple[User, Result[APIUser, pydantic_core.ValidationError]]:
        request = RequestFactory().get("/")
        user = User(
            username=str(uuid.uuid4()),
            first_name="joe" if valid else 42,
            last_name="schmoe",
        )
        return user, APIUser.from_django(request, user)

    def test_accept_model(self) -> None:
        user, maybeobj = self._get_api_user(valid=True)
        assert maybeobj.is_ok
        response = APIResponse(maybeobj.ok())
        self.assertEqual(response.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            response.content,
            {
                "username": user.username,
                "first_name": "joe",
                "last_name": "schmoe",
            },
        )

    def test_accept_result_containing_model(self) -> None:
        user, maybeobj = self._get_api_user(valid=True)
        response = APIResponse(maybeobj)
        self.assertEqual(response.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            response.content,
            {
                "username": user.username,
                "first_name": "joe",
                "last_name": "schmoe",
            },
        )

    def test_accept_result_containing_errors(self) -> None:
        user, maybeobj = self._get_api_user(valid=False)
        response = APIResponse(maybeobj)
        self.assertEqual(response.status_code, Status.HTTP_400_BAD_REQUEST)
        self.assertJSONEqual(
            response.content,
            {
                "errors": {
                    "first_name": {
                        "type": "string_type",
                        "msg": "Input should be a valid string",
                    }
                }
            },
        )



from typing import assert_type
from unittest import TestCase

from thelabtyping.result import (
    Err,
    Ok,
    Result,
    UnwrapError,
    as_result,
    do,
    partition_results,
)


class OkTest(TestCase):
    def test_ok_result(self) -> None:
        res = Ok("foo")
        self.assertTrue(res.is_ok)
        self.assertFalse(res.is_err)
        self.assertEqual(res.ok(), "foo")
        self.assertEqual(res.err(), None)  # type:ignore[func-returns-value]
        self.assertEqual(res.unwrap(), "foo")
        with self.assertRaises(UnwrapError):
            res.unwrap_err()

    def test_dunders(self) -> None:
        self.assertEqual(Ok(5), Ok(5))
        self.assertNotEqual(Ok(5), Ok(6))
        self.assertNotEqual(Ok(5), Err(5))
        self.assertIsNotNone(hash(Ok(1)))
        self.assertEqual(repr(Ok(1)), "Ok(1)")


class ErrTest(TestCase):
    def test_err_result(self) -> None:
        res = Err("foo")
        self.assertFalse(res.is_ok)
        self.assertTrue(res.is_err)
        self.assertEqual(res.ok(), None)  # type:ignore[func-returns-value]
        self.assertEqual(res.err(), "foo")
        with self.assertRaises(UnwrapError):
            res.unwrap()
        self.assertEqual(res.unwrap_err(), "foo")

    def test_dunders(self) -> None:
        self.assertEqual(Err(5), Err(5))
        self.assertNotEqual(Err(5), Err(6))
        self.assertNotEqual(Err(5), Ok(5))
        self.assertIsNotNone(hash(Err(1)))
        self.assertEqual(repr(Err(1)), "Err(1)")


class ResultTest(TestCase):
    def test_result(self) -> None:
        def foo(x: int) -> Result[int, str]:
            if x > 5:
                return Ok(x)
            return Err("x is too small")

        self.assertIsInstance(foo(1), Err)
        self.assertIsInstance(foo(6), Ok)

    def test_as_result(self) -> None:
        @as_result(ValueError)
        def foo(x: int) -> int:
            if x > 5:
                return x
            raise ValueError("x is too small")

        resA = foo(1)
        assert_type(resA, Result[int, ValueError])
        if resA.is_ok:
            assert_type(resA, Ok[int])
        else:
            assert_type(resA, Err[ValueError])

        self.assertIsInstance(foo(1), Err)
        self.assertIsInstance(foo(6), Ok)

    def test_as_result_multi_catch_types(self) -> None:
        @as_result(ValueError, TypeError)
        def foo(x: int) -> int:
            if x > 5:
                return x
            raise ValueError("x is too small")

        resA = foo(1)
        assert_type(resA, Result[int, Exception])
        if resA.is_ok:
            assert_type(resA, Ok[int])
        else:
            assert_type(resA, Err[Exception])

        self.assertIsInstance(foo(1), Err)
        self.assertIsInstance(foo(6), Ok)

    def test_and_then(self) -> None:
        def foo(x: int) -> Result[int, str]:
            if x > 5:
                return Ok(x)
            return Err("x is too small")

        @as_result(Exception)
        def add_five(x: int) -> int:
            return x + 5

        resA = foo(1).and_then(add_five)
        assert_type(resA, Ok[int] | Err[str] | Err[Exception])
        self.assertTrue(resA.is_err)
        self.assertEqual(resA.err(), "x is too small")

        resB = foo(6).and_then(add_five)
        assert_type(resB, Ok[int] | Err[str] | Err[Exception])
        self.assertTrue(resB.is_ok)
        self.assertEqual(resB.ok(), 11)

    def test_do_notation(self) -> None:
        def foo(x: int) -> Result[int, str]:
            if x > 5:
                return Ok(x)
            return Err("x is too small")

        resA: Result[int, str] = do(Ok(x + 5) for x in foo(1))
        assert_type(resA, Result[int, str])
        self.assertEqual(resA.err(), "x is too small")

        resB: Result[int, str] = do(Ok(x + 5) for x in foo(6))
        assert_type(resB, Result[int, str])
        self.assertEqual(resB.ok(), 11)

        # Two calls to foo, both succeed
        resC: Result[int, str] = do(Ok(x + y) for x in foo(6) for y in foo(6))
        assert_type(resC, Result[int, str])
        self.assertEqual(resC.ok(), 12)

        # Error in y
        resD: Result[int, str] = do(Ok(x + y) for x in foo(6) for y in foo(1))
        assert_type(resD, Result[int, str])
        self.assertEqual(resD.err(), "x is too small")

    def test_partition_results(self) -> None:
        mixed_results: list[Result[int, str]] = [
            Ok(1),
            Ok(2),
            Err("foo"),
            Ok(3),
            Err("bar"),
        ]
        oks, errs = partition_results(mixed_results)
        assert_type(oks, list[int])
        assert_type(errs, list[str])
        self.assertEqual(
            oks,
            [
                1,
                2,
                3,
            ],
        )
        self.assertEqual(
            errs,
            [
                "foo",
                "bar",
            ],
        )



"""OpenAPI 3.1 specification generation for thelab-typing Router."""

from dataclasses import dataclass, field
from typing import cast
import hashlib

from django.core.cache import cache
from django.http import HttpRequest, HttpResponse, HttpResponseForbidden

from ..responses import APIResponse
from ..routing import Router
from .generator import OpenAPIGenerator, OpenAPIGeneratorConfig
from .models import OpenAPI, Server


@dataclass
class OpenAPISchema:
    """OpenAPI 3.1 schema generator that implements SchemaGenerator protocol."""

    title: str
    version: str
    description: str | None = None
    servers: list[Server] = field(default_factory=list)
    cache_timeout: int | None = 3600
    require_auth: bool = True

    def __call__(self, router: "Router", request: HttpRequest) -> HttpResponse:
        """Handle schema request - implements SchemaGenerator protocol."""
        # Check authentication first
        if self.require_auth:
            if (
                not hasattr(request, "user")
                or not request.user
                or request.user.is_anonymous
            ):
                return HttpResponseForbidden()

        # Auto-populate servers from request if not provided
        servers = self.servers
        if not servers:
            # Build server URL from request (scheme + host)
            base_url = request.build_absolute_uri("/").rstrip("/")
            servers = [Server(url=base_url)]

        # Get namespace from request for full path resolution
        namespace = request.resolver_match.namespace if request.resolver_match else None

        # Try to get cached response
        def _inner() -> OpenAPI:
            config = OpenAPIGeneratorConfig(
                title=self.title,
                version=self.version,
                description=self.description,
                servers=servers,
            )
            generator = OpenAPIGenerator(config)
            return generator.generate_from_routes(router.routes, namespace=namespace)

        cache_key = self._get_cache_key(router, servers, namespace)
        spec = cast(
            OpenAPI,
            cache.get_or_set(cache_key, _inner, timeout=self.cache_timeout)
            if self.cache_timeout is not None
            else _inner(),
        )
        return APIResponse(spec, exclude_none=True)

    def _get_cache_key(
        self,
        router: "Router",
        servers: list[Server],
        namespace: str | None,
    ) -> str:
        """Generate a cache key for the OpenAPI spec."""
        # Include all factors that affect the generated spec
        key_parts = [
            "openapi_schema",
            self.title,
            self.version,
            self.description or "",
            ",".join(s.url for s in servers),
            namespace or "",
            router.basename or "",
        ]
        key_data = "|".join(key_parts)
        return f"openapi:{hashlib.md5(key_data.encode(), usedforsecurity=False).hexdigest()}"



from django.contrib.auth.models import AnonymousUser, User
from django.http import HttpRequest
import pydantic

from .serializers import Empty

"""
Alias for any user: e.g. maybe authenticated, maybe not.
"""
type AnyUser = User | AnonymousUser


class TypedRequest[
    _QST: pydantic.BaseModel,  # QueryString Type
    _BT: pydantic.BaseModel,  # Body Type
    _UT: AnyUser = AnyUser,  # User Type
](HttpRequest):
    """
    Wrapper around the stock HttpRequest class. Provides typed-checked,
    validated (via Pydantic) data models for the querystring, the request body,
    or both. Can also assert (and type) that the user is authenticated.
    """

    user: _UT

    __validated_querystring: _QST
    __validated_body: _BT

    def __init__(self, request: HttpRequest, user: _UT, qs: _QST, body: _BT) -> None:
        # Clone the original request into self
        self.__dict__.update(request.__dict__)
        # Add the validate request data objects
        self.user = user
        self.__validated_querystring = qs
        self.__validated_body = body

    @property
    def validated_querystring(self) -> _QST:
        """
        Return the fully-validated querystring data. This is a property merely
        to avoid accidental mutation.
        """
        return self.__validated_querystring

    @property
    def validated_body(self) -> _BT:
        """
        Return the fully-validated body (POST/PUT/etc) data. This is a property
        merely to avoid accidental mutation.
        """
        return self.__validated_body


"""
Alias for defining a TypedRequest with a validated querystring and an empty
body.
"""
type TypedRequestQuery[
    _QST: pydantic.BaseModel,
    _UT: AnyUser = AnyUser,
] = TypedRequest[_QST, Empty, _UT]

"""
Alias for defining a TypedRequest with an empty querystring and an validated
body.
"""
type TypedRequestBody[
    _BT: pydantic.BaseModel,
    _UT: AnyUser = AnyUser,
] = TypedRequest[Empty, _BT, _UT]

"""
Alias for defining a TypedRequest with an empty querystring and an empty body.
"""
type EmptyTypedRequest[
    _UT: AnyUser = AnyUser,
] = TypedRequest[Empty, Empty, _UT]


"""
Alias for defining a TypedRequest with a empty querystring, a validated body,
and and authenticated user.
"""
type AuthdTypedRequest[
    _QST: pydantic.BaseModel,
    _BT: pydantic.BaseModel,
    _UT: User = User,
] = TypedRequest[_QST, _BT, _UT]

"""
Alias for defining a TypedRequest with a validated querystring, an empty body,
and and authenticated user.
"""
type AuthdTypedRequestQuery[
    _QST: pydantic.BaseModel,
    _UT: User = User,
] = TypedRequest[_QST, Empty, _UT]

"""
Alias for defining a TypedRequest with a empty querystring, a validated body,
and and authenticated user.
"""
type AuthdTypedRequestBody[
    _BT: pydantic.BaseModel,
    _UT: User = User,
] = TypedRequest[Empty, _BT, _UT]

"""
Alias for defining a TypedRequest with a empty querystring, an empty body,
and and authenticated user.
"""
type AuthdEmptyTypedRequest[
    _UT: User = User,
] = TypedRequest[Empty, Empty, _UT]



"""OpenAPI specification generator from Router."""

from dataclasses import dataclass, field
from typing import Any
import logging
import re

from django.http import HttpResponseBase
from django.urls import reverse
import pydantic

from ..routing import HttpMethod, RegisteredView, Route, RouteMap, Router
from ..status import Status
from .introspection import ViewTypeInfo, infer_view_types, is_empty_model
from .models import (
    Components,
    Info,
    MediaType,
    OpenAPI,
    Operation,
    Parameter,
    PathItem,
    Reference,
    RequestBody,
    Response,
    Server,
)
from .path_parser import (
    PathParam,
    get_placeholder_kwargs,
    parse_django_path_params,
    replace_placeholders_with_openapi_params,
)

logger = logging.getLogger(__name__)


@dataclass
class OpenAPIGeneratorConfig:
    """Configuration for OpenAPI spec generation."""

    title: str
    version: str
    description: str | None = None
    servers: list[Server] = field(default_factory=list)
    auth_security_scheme: str | None = None
    default_success_status: Status = Status.HTTP_200_OK
    include_validation_errors: bool = True


class OpenAPIGenerator:
    """Generates OpenAPI 3.1 specifications from Router instances."""

    def __init__(
        self,
        config: OpenAPIGeneratorConfig,
    ) -> None:
        self.config = config
        self._schemas: dict[str, dict[str, Any]] = {}
        self._schema_refs: dict[str, str] = {}

    def generate(
        self,
        router: Router,
        namespace: str | None = None,
    ) -> OpenAPI:
        """Generate a complete OpenAPI spec from a Router."""
        return self.generate_from_routes(router.routes, namespace)

    def generate_from_routes(
        self,
        routes: RouteMap,
        namespace: str | None = None,
    ) -> OpenAPI:
        """Generate a complete OpenAPI spec from a RouteMap."""
        # Reset schema cache for this generation
        self._schemas = {}
        self._schema_refs = {}

        # Build paths
        paths: dict[str, PathItem] = {}
        for pattern, route in routes.items():
            # Resolve full path using reverse() if namespace is available
            openapi_path = self._resolve_openapi_path(pattern, route, namespace)
            path_item = self._build_path_item(pattern, route)
            paths[openapi_path] = path_item

        # Build components
        components = None
        if self._schemas:
            components = Components(schemas=self._schemas)

        return OpenAPI(
            info=Info(
                title=self.config.title,
                version=self.config.version,
                description=self.config.description,
            ),
            servers=self.config.servers if self.config.servers else None,
            paths=paths,
            components=components,
        )

    def _resolve_openapi_path(
        self,
        pattern: str,
        route: Route,
        namespace: str | None = None,
    ) -> str:
        """Resolve the full OpenAPI path for a route."""
        # Get placeholder kwargs for any path parameters
        placeholder_kwargs = get_placeholder_kwargs(pattern)
        qualified_name = f"{namespace}:{route.name}" if namespace else route.name
        full_url = reverse(qualified_name, kwargs=placeholder_kwargs or None)
        return replace_placeholders_with_openapi_params(full_url, placeholder_kwargs)

    def _build_path_item(self, pattern: str, route: Route) -> PathItem:
        """Build a PathItem from a Route."""
        path_params = parse_django_path_params(pattern)

        operations: dict[str, Operation | None] = {
            "get": None,
            "post": None,
            "put": None,
            "patch": None,
            "delete": None,
        }

        for view in route.views:
            for method in view.methods:
                operation = self._build_operation(method, view, path_params)
                method_name = method.value.lower()
                operations[method_name] = operation

        return PathItem(
            get=operations["get"],
            post=operations["post"],
            put=operations["put"],
            patch=operations["patch"],
            delete=operations["delete"],
        )

    def _build_operation(
        self,
        method: HttpMethod,
        view: RegisteredView[[], HttpResponseBase],
        path_params: list[PathParam],
    ) -> Operation:
        """Build an Operation from a RegisteredView."""
        view_info = infer_view_types(view.fn)

        # Build parameters (path + query)
        parameters = self._build_parameters(view_info, path_params)

        # Build request body
        request_body = None
        if view_info.body_model is not None:
            request_body = self._build_request_body(view_info.body_model)

        # Build responses
        responses = self._build_responses(view_info.response_model)

        # Security
        security: list[dict[str, list[str]]] | None = None
        if view_info.requires_auth and self.config.auth_security_scheme:
            security = [{self.config.auth_security_scheme: []}]

        # Operation ID from function name
        operation_id = view.fn.__name__

        params_list: list[Parameter | Reference] | None = (
            list(parameters) if parameters else None
        )
        responses_dict: dict[str, Response | Reference] = dict(responses)

        return Operation(
            operationId=operation_id,
            parameters=params_list,
            requestBody=request_body,
            responses=responses_dict,
            security=security,
        )

    def _build_parameters(
        self,
        view_info: ViewTypeInfo,
        path_params: list[PathParam],
    ) -> list[Parameter]:
        """Build Parameter objects from query model and path params."""
        parameters: list[Parameter] = []

        # Path parameters
        for param in path_params:
            schema: dict[str, Any] = {"type": param.openapi_type}
            if param.openapi_format:
                schema["format"] = param.openapi_format

            parameters.append(
                Parameter(
                    name=param.name,
                    param_in="path",
                    required=True,
                    param_schema=schema,
                )
            )

        # Query parameters from model
        if view_info.querystring_model is not None:
            # Get the JSON schema for the query model
            # Note: We call _get_or_create_schema_ref for side effects (populates _schemas)
            # but query params are expanded inline, not as a $ref
            self._get_or_create_schema_ref(view_info.querystring_model)
            # For query parameters, we reference the schema
            # Individual fields become query params
            json_schema = view_info.querystring_model.model_json_schema()
            properties = json_schema.get("properties", {})
            required_fields = set(json_schema.get("required", []))

            for field_name, field_schema in properties.items():
                parameters.append(
                    Parameter(
                        name=field_name,
                        param_in="query",
                        required=field_name in required_fields,
                        param_schema=field_schema,
                    )
                )

        return parameters

    def _build_request_body(self, body_model: type[pydantic.BaseModel]) -> RequestBody:
        """Build RequestBody from body model."""
        ref = self._get_or_create_schema_ref(body_model)

        return RequestBody(
            content={
                "application/json": MediaType(
                    media_type_schema=ref,
                ),
            },
            required=True,
        )

    def _build_responses(
        self,
        response_model: type[pydantic.BaseModel] | None,
    ) -> dict[str, Response]:
        """Build Response objects from response model."""
        responses: dict[str, Response] = {}
        success_status_key = str(self.config.default_success_status.value)

        if response_model is not None and not is_empty_model(response_model):
            ref = self._get_or_create_schema_ref(response_model)
            responses[success_status_key] = Response(
                description="Successful response",
                content={
                    "application/json": MediaType(
                        media_type_schema=ref,
                    ),
                },
            )
        else:
            responses[success_status_key] = Response(
                description="Successful response",
            )

        # Add validation error response
        if self.config.include_validation_errors:
            responses[str(Status.HTTP_400_BAD_REQUEST.value)] = Response(
                description="Validation error",
            )

        return responses

    def _get_or_create_schema_ref(
        self,
        model: type[pydantic.BaseModel],
    ) -> Reference:
        """Get or create a component schema reference for a Pydantic model."""
        # Get a unique name for this model
        model_name = self._sanitize_schema_name(self._get_model_name(model))

        if model_name in self._schema_refs:
            return Reference(ref=self._schema_refs[model_name])

        # Generate JSON Schema
        try:
            schema = model.model_json_schema(
                ref_template="#/components/schemas/{model}",
                mode="serialization",
            )
        except (
            pydantic.PydanticSchemaGenerationError,
            pydantic.PydanticUserError,
        ) as e:
            # Fallback for generic types that can't generate schema
            logger.warning(
                "Failed to generate JSON schema for model %s: %s. Using fallback.",
                model_name,
                e,
            )
            schema = {"type": "object"}

        # Extract $defs if present and merge into components
        # Sanitize all $def names to comply with OpenAPI schema naming rules
        defs = schema.pop("$defs", {})
        for def_name, def_schema in defs.items():
            sanitized_name = self._sanitize_schema_name(def_name)
            if sanitized_name not in self._schemas:
                # Update any internal $ref references to use sanitized names
                sanitized_schema = self._sanitize_schema_refs(def_schema)
                self._schemas[sanitized_name] = sanitized_schema

        # Handle allOf references that point to $defs
        if "allOf" in schema and len(schema["allOf"]) == 1:
            ref_obj = schema["allOf"][0]
            if "$ref" in ref_obj:
                # The schema is just a reference to a $def
                ref_name = ref_obj["$ref"].split("/")[-1]
                sanitized_ref_name = self._sanitize_schema_name(ref_name)
                if sanitized_ref_name in self._schemas:
                    # Use the existing schema name
                    ref_path = f"#/components/schemas/{sanitized_ref_name}"
                    self._schema_refs[model_name] = ref_path
                    return Reference(ref=ref_path)

        # Sanitize any refs in the main schema
        schema = self._sanitize_schema_refs(schema)

        # Store the main schema
        self._schemas[model_name] = schema

        ref_path = f"#/components/schemas/{model_name}"
        self._schema_refs[model_name] = ref_path

        return Reference(ref=ref_path)

    def _get_model_name(self, model: type[pydantic.BaseModel]) -> str:
        """Get a unique name for a Pydantic model."""
        if hasattr(model, "__name__"):
            return model.__name__

        # For generic types like ListOf[T], create a descriptive name
        return str(model)

    def _sanitize_schema_name(self, name: str) -> str:
        """Sanitize a schema name to comply with OpenAPI naming rules."""
        # Replace any character not in the allowed set with underscore
        result = re.sub(r"[^a-zA-Z0-9._-]", "_", name)
        # Remove any trailing underscores
        result = result.rstrip("_")
        # Ensure we have a valid name (fallback for edge cases)
        if not result:
            return "UnnamedSchema"
        return result

    def _sanitize_schema_refs(self, schema: dict[str, Any]) -> dict[str, Any]:
        """Recursively sanitize $ref paths in a schema to use sanitized names."""
        if not isinstance(schema, dict):
            return schema

        result: dict[str, Any] = {}
        for key, value in schema.items():
            if key == "$ref" and isinstance(value, str):
                # Extract and sanitize the reference name
                if value.startswith("#/components/schemas/"):
                    ref_name = value.split("/")[-1]
                    sanitized_name = self._sanitize_schema_name(ref_name)
                    result[key] = f"#/components/schemas/{sanitized_name}"
                else:
                    result[key] = value
            elif isinstance(value, dict):
                result[key] = self._sanitize_schema_refs(value)
            elif isinstance(value, list):
                result[key] = [
                    self._sanitize_schema_refs(item) if isinstance(item, dict) else item
                    for item in value
                ]
            else:
                result[key] = value

        return result


def generate_openapi(
    router: Router,
    title: str,
    version: str,
    description: str | None = None,
    servers: list[Server] | None = None,
    namespace: str | None = None,
) -> OpenAPI:
    """Generate OpenAPI 3.1 spec from a Router with minimal configuration."""
    config = OpenAPIGeneratorConfig(
        title=title,
        version=version,
        description=description,
        servers=servers or [],
    )
    generator = OpenAPIGenerator(config)
    return generator.generate(router, namespace)



"""Type introspection utilities for extracting OpenAPI schema info from views."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, get_args, get_origin, get_type_hints
import logging

import pydantic

from ..requests import (
    AuthdEmptyTypedRequest,
    AuthdTypedRequest,
    AuthdTypedRequestBody,
    AuthdTypedRequestQuery,
    EmptyTypedRequest,
    TypedRequest,
    TypedRequestBody,
    TypedRequestQuery,
)
from ..responses import APIResponse
from ..serializers import Empty

logger = logging.getLogger(__name__)


def _safe_cast_to_model(
    value: Any,
) -> type[pydantic.BaseModel] | None:
    """Safely cast a type argument to a Pydantic BaseModel subclass."""
    if value is None:
        return None
    if isinstance(value, type) and issubclass(value, pydantic.BaseModel):
        return value
    logger.warning(
        "Expected BaseModel subclass, got %s. Ignoring type argument.",
        type(value).__name__,
    )
    return None


@dataclass
class ViewTypeInfo:
    """Extracted type information from a view function."""

    querystring_model: type[pydantic.BaseModel] | None
    body_model: type[pydantic.BaseModel] | None
    response_model: type[pydantic.BaseModel] | None
    requires_auth: bool


def is_empty_model(model: type[pydantic.BaseModel] | None) -> bool:
    """Check if a model is the Empty sentinel model or None."""
    return model is None or model is Empty


def _normalize_model(
    model: type[pydantic.BaseModel] | None,
) -> type[pydantic.BaseModel] | None:
    """Return None if model is Empty, otherwise return the model."""
    if is_empty_model(model):
        return None
    return model


def _infer_request_types(
    request_type: type[Any] | None,
) -> tuple[type[pydantic.BaseModel] | None, type[pydantic.BaseModel] | None, bool]:
    """Extract query model, body model, and auth requirement from request type."""
    if request_type is None:
        return None, None, False

    origin = get_origin(request_type)
    if origin is None:
        origin = request_type
    args = get_args(request_type)

    # Auth-required types
    if origin is AuthdEmptyTypedRequest:
        return None, None, True

    if origin is AuthdTypedRequestQuery:
        qs_model = _safe_cast_to_model(args[0]) if args else None
        return _normalize_model(qs_model), None, True

    if origin is AuthdTypedRequestBody:
        body_model = _safe_cast_to_model(args[0]) if args else None
        return None, _normalize_model(body_model), True

    if origin is AuthdTypedRequest:
        qs_model = _safe_cast_to_model(args[0]) if len(args) > 0 else None
        body_model = _safe_cast_to_model(args[1]) if len(args) > 1 else None
        return _normalize_model(qs_model), _normalize_model(body_model), True

    # Auth-not-required types
    if origin is EmptyTypedRequest:
        return None, None, False

    if origin is TypedRequestQuery:
        qs_model = _safe_cast_to_model(args[0]) if args else None
        return _normalize_model(qs_model), None, False

    if origin is TypedRequestBody:
        body_model = _safe_cast_to_model(args[0]) if args else None
        return None, _normalize_model(body_model), False

    if origin is TypedRequest:
        qs_model = _safe_cast_to_model(args[0]) if len(args) > 0 else None
        body_model = _safe_cast_to_model(args[1]) if len(args) > 1 else None
        return _normalize_model(qs_model), _normalize_model(body_model), False

    # Unknown request type - return defaults
    return None, None, False


def _extract_response_model(
    return_type: type[Any] | None,
) -> type[pydantic.BaseModel] | None:
    """Extract the model type T from APIResponse[T]."""
    if return_type is None:
        return None

    origin = get_origin(return_type)
    if origin is not APIResponse:
        return None

    args = get_args(return_type)
    if not args:
        return None

    response_type = args[0]

    # Handle generic types like ListOf[T] which are subscripted generics
    response_origin = get_origin(response_type)
    if response_origin is not None:
        # It's a generic like ListOf[T] - return the subscripted generic itself
        # This preserves type info for schema generation
        if hasattr(response_type, "__mro__") or hasattr(response_origin, "__mro__"):
            # Check if it's Empty
            inner_args = get_args(response_type)
            if inner_args and inner_args[0] is Empty:
                return None
            # Return the generic origin class (e.g., ListOf)
            # The actual type info will be handled by Pydantic's model_json_schema
            return response_type  # type: ignore[no-any-return]

    # Check if it's a concrete Pydantic model
    if isinstance(response_type, type) and issubclass(
        response_type, pydantic.BaseModel
    ):
        if is_empty_model(response_type):
            return None
        return response_type

    return None


def infer_view_types(fn: Callable[..., Any]) -> ViewTypeInfo:
    """Extract type information from a view function's annotations."""
    try:
        hints = get_type_hints(fn)
    except (NameError, AttributeError, TypeError) as e:
        # Type hint resolution can fail due to:
        # - NameError: Forward references that can't be resolved
        # - AttributeError: Invalid type annotations
        # - TypeError: Invalid type hint syntax
        logger.warning(
            "Failed to resolve type hints for function %s: %s. "
            "OpenAPI schema will be incomplete.",
            getattr(fn, "__name__", fn),
            e,
        )
        return ViewTypeInfo(
            querystring_model=None,
            body_model=None,
            response_model=None,
            requires_auth=False,
        )

    request_type = hints.get("request")
    return_type = hints.get("return")

    qs_model, body_model, requires_auth = _infer_request_types(request_type)
    response_model = _extract_response_model(return_type)

    return ViewTypeInfo(
        querystring_model=qs_model,
        body_model=body_model,
        response_model=response_model,
        requires_auth=requires_auth,
    )



from typing import Literal, assert_type

from django.contrib.auth.models import AnonymousUser, User
from django.http import HttpRequest
from django.test import RequestFactory, TestCase

from thelabtyping.api.requests import (
    AuthdEmptyTypedRequest,
    AuthdTypedRequest,
    AuthdTypedRequestBody,
    AuthdTypedRequestQuery,
    EmptyTypedRequest,
    TypedRequest,
    TypedRequestBody,
)
from thelabtyping.api.responses import APIResponse
from thelabtyping.api.serializers import Empty
from thelabtyping.api.status import Status
from thelabtyping.api.views import validate

from ..sampleapp.serializers import APIUser, UserSearchQuery


class FunctionalViewTest(TestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(username="joe", password="password")

    def test_typed_request(self) -> None:
        @validate()
        def view(
            request: TypedRequest[UserSearchQuery, APIUser],
        ) -> APIResponse[Empty]:
            # Some assertions to prove the `validate` decorator is doing what it's
            # supposed to
            assert_type(request, TypedRequest[UserSearchQuery, APIUser])
            assert_type(request.user, User | AnonymousUser)
            assert_type(request.validated_querystring, UserSearchQuery)
            assert_type(request.validated_body, APIUser)
            return APIResponse(Empty())

        req = RequestFactory().post(
            "/",
            data={
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )
        resp = view(req)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)

    def test_typed_request_body(self) -> None:
        @validate()
        def view(
            request: TypedRequestBody[APIUser],
        ) -> APIResponse[Empty]:
            # Some assertions to prove the `validate` decorator is doing what it's
            # supposed to
            assert_type(request, TypedRequest[Empty, APIUser])
            assert_type(request.user, User | AnonymousUser)
            assert_type(request.validated_querystring, Empty)
            assert_type(request.validated_body, APIUser)
            return APIResponse(Empty())

        req = RequestFactory().post(
            "/",
            data={
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )
        resp = view(req)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)

    def test_empty_request(self) -> None:
        @validate()
        def view(
            request: EmptyTypedRequest,
        ) -> APIResponse[Empty]:
            # Some assertions to prove the `validate` decorator is doing what it's
            # supposed to
            assert_type(request, TypedRequest[Empty, Empty])
            assert_type(request.user, User | AnonymousUser)
            assert_type(request.validated_querystring, Empty)
            assert_type(request.validated_body, Empty)
            return APIResponse(Empty())

        req = RequestFactory().get("/")
        resp = view(req)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)

    def test_authd_typed_request(self) -> None:
        @validate()
        def view(
            request: AuthdTypedRequest[UserSearchQuery, APIUser],
        ) -> APIResponse[Empty]:
            # Some assertions to prove the `validate` decorator is doing what it's
            # supposed to
            assert_type(request, TypedRequest[UserSearchQuery, APIUser, User])
            assert_type(request.user, User)
            assert_type(request.user.is_authenticated, Literal[True])
            assert_type(request.user.is_anonymous, Literal[False])
            assert_type(request.validated_querystring, UserSearchQuery)
            assert_type(request.validated_body, APIUser)
            return APIResponse(Empty())

        req = RequestFactory().post(
            "/",
            data={
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )
        req.user = self.user
        resp = view(req)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)

    def test_authd_typed_request_query(self) -> None:
        @validate()
        def view(
            request: AuthdTypedRequestQuery[UserSearchQuery],
        ) -> APIResponse[Empty]:
            # Some assertions to prove the `validate` decorator is doing what it's
            # supposed to
            assert_type(request, TypedRequest[UserSearchQuery, Empty, User])
            assert_type(request.user, User)
            assert_type(request.user.is_authenticated, Literal[True])
            assert_type(request.user.is_anonymous, Literal[False])
            assert_type(request.validated_querystring, UserSearchQuery)
            assert_type(request.validated_body, Empty)
            return APIResponse(Empty())

        req = RequestFactory().get("/")
        req.user = self.user
        resp = view(req)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)

    def test_authd_typed_request_body(self) -> None:
        @validate()
        def view(
            request: AuthdTypedRequestBody[APIUser],
        ) -> APIResponse[Empty]:
            # Some assertions to prove the `validate` decorator is doing what it's
            # supposed to
            assert_type(request, TypedRequest[Empty, APIUser, User])
            assert_type(request.user, User)
            assert_type(request.user.is_authenticated, Literal[True])
            assert_type(request.user.is_anonymous, Literal[False])
            assert_type(request.validated_querystring, Empty)
            assert_type(request.validated_body, APIUser)
            return APIResponse(Empty())

        req = RequestFactory().post(
            "/",
            data={
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )
        req.user = self.user
        resp = view(req)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)

    def test_authd_typed_request_empty(self) -> None:
        @validate()
        def view(
            request: AuthdEmptyTypedRequest,
        ) -> APIResponse[Empty]:
            # Some assertions to prove the `validate` decorator is doing what it's
            # supposed to
            assert_type(request, TypedRequest[Empty, Empty, User])
            assert_type(request.user, User)
            assert_type(request.user.is_authenticated, Literal[True])
            assert_type(request.user.is_anonymous, Literal[False])
            assert_type(request.validated_querystring, Empty)
            assert_type(request.validated_body, Empty)
            return APIResponse(Empty())

        req = RequestFactory().get("/")
        req.user = self.user
        resp = view(req)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)

    def test_invalid_request_type(self) -> None:
        with self.assertRaises(TypeError):

            @validate()
            def view(
                request: HttpRequest,
            ) -> APIResponse[Empty]:
                # Some assertions to prove the `validate` decorator is doing what it's
                # supposed to
                assert_type(request, HttpRequest)
                assert_type(request.user, User | AnonymousUser)
                return APIResponse(Empty())



import os

import django_stubs_ext

django_stubs_ext.monkeypatch()

ROOT_URLCONF = "thelabtyping.tests.settings.urls"

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"
INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "rest_framework",
    "thelabtyping.tests.sampleapp",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

SECRET_KEY = "secret"

HERE = os.path.dirname(os.path.abspath(__file__))
DB = os.path.join(HERE, "testdb.sqlite")

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": DB,
        "TEST": {
            "NAME": DB,
        },
    },
}



from django.urls import include, path

from thelabtyping.api.routing import Router

from .views import (
    UserView,
    router,
    router_users_detail,
    router_users_list,
    router_users_post,
)

# Alternate method of building a router
alt_router = Router(basename="alt-router", enable_index=False)
alt_router.route(
    "users/",
    name="users-list",
    get=router_users_list,
    post=router_users_post,
)
alt_router.route(
    "users/<int:pk>/",
    name="users-detail",
    get=router_users_detail,
)

app_name = "sampleapp"
urlpatterns = [
    path("users/", UserView.as_view(), name="users-list"),
    path("router/", include(router.urls)),
    path("alt-router/", include(alt_router.urls)),
]



from collections.abc import Mapping
from typing import Any
import json

from pydantic import BaseModel
from rest_framework import renderers
from rest_framework.compat import INDENT_SEPARATORS, LONG_SEPARATORS, SHORT_SEPARATORS


class JSONRenderer(renderers.JSONRenderer):
    """
    Same as the normal JSONRenderer, but knows how to work with Pydantic models.
    """

    def render(
        self,
        data: BaseModel | Any,
        accepted_media_type: str | None = None,
        renderer_context: Mapping[str, Any] | None = None,
    ) -> bytes:
        """
        Render `data` into JSON, returning a bytestring.
        """
        if data is None:
            return b""

        renderer_context = renderer_context or {}
        indent = self.get_indent(accepted_media_type or "", renderer_context or {})
        if indent is None:
            separators = SHORT_SEPARATORS if self.compact else LONG_SEPARATORS
        else:
            separators = INDENT_SEPARATORS

        if isinstance(data, BaseModel):
            ret = data.model_dump_json(
                indent=indent,
            )
        else:
            ret = json.dumps(
                data,
                cls=self.encoder_class,
                indent=indent,
                ensure_ascii=self.ensure_ascii,
                allow_nan=not self.strict,
                separators=separators,
            )

        # We always fully escape \u2028 and \u2029 to ensure we output JSON
        # that is a strict javascript subset.
        # See: https://gist.github.com/damncabbage/623b879af56f850a6ddc
        ret = ret.replace("\u2028", "\\u2028").replace("\u2029", "\\u2029")
        return ret.encode()


class BrowsableAPIRenderer(renderers.BrowsableAPIRenderer):
    """
    Same as the normal BrowsableAPIRenderer, but knows how to work with Pydantic models.
    """

    def render(
        self,
        data: BaseModel | Any,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """Render data for the browsable API, converting Pydantic models to dicts."""
        if isinstance(data, BaseModel):
            data = data.model_dump()
        return super().render(data, *args, **kwargs)



from typing import Self

from django.contrib.auth.models import User
from django.db.models import QuerySet
from django.http import HttpRequest
import pydantic
import pydantic_core

from thelabtyping.abc import ListOf
from thelabtyping.api.serializers import APIObj
from thelabtyping.result import Ok, Result, as_result


class UserSearchQuery(pydantic.BaseModel):
    id: int | None = None
    first_name: str | None = None

    def get_queryset(self) -> QuerySet[User]:
        qs = User.objects.all()
        if self.id:
            qs = qs.filter(id=self.id)
        if self.first_name:
            qs = qs.filter(first_name=self.first_name)
        return qs


class APIUser(APIObj[User]):
    """
    Example of an API serializer for the User model
    """

    username: str
    first_name: str
    last_name: str

    @classmethod
    @as_result(pydantic_core.ValidationError)
    def from_django(cls, request: HttpRequest, obj: User) -> Self:
        return cls(
            username=obj.username,
            first_name=obj.first_name,
            last_name=obj.last_name,
        )

    def create(self, request: HttpRequest) -> Result[User, Exception]:
        user = User.objects.create_user(
            username=self.username,
            first_name=self.first_name,
            last_name=self.last_name,
        )
        return Ok(user)


class APIUserWithCustomList(APIUser):
    class List(ListOf["APIUserWithCustomList"]):
        pass



"""Shared test fixtures for OpenAPI tests."""

from typing import Any
import json

from django.test import override_settings
from django.urls import include, path
import jsonschema
import pydantic

from thelabtyping.api.openapi.generator import (
    OpenAPIGenerator,
    OpenAPIGeneratorConfig,
    generate_openapi,
)
from thelabtyping.api.openapi.models import OpenAPI, Server
from thelabtyping.api.routing import Router


class QueryModel(pydantic.BaseModel):
    """Sample query model for tests."""

    search: str | None = None
    limit: int = 10


class BodyModel(pydantic.BaseModel):
    """Sample body model for tests."""

    name: str
    email: str


class ResponseModel(pydantic.BaseModel):
    """Sample response model for tests."""

    id: int
    name: str


class OpenAPITestMixin:
    """Mixin for tests that need to generate OpenAPI specs from test routers.

    Dynamically registers test routers in Django's URL config so that
    reverse() works correctly during OpenAPI generation.
    """

    def generate_openapi_for_test(
        self,
        router: Router,
        title: str = "Test API",
        version: str = "1.0.0",
        description: str | None = None,
        servers: list[Server] | None = None,
        namespace: str = "test",
    ) -> OpenAPI:
        """Generate OpenAPI spec after registering router in URL config."""

        class TestURLConf:
            urlpatterns = [
                path("test/", include((router.urls, namespace))),
            ]

        with override_settings(ROOT_URLCONF=TestURLConf):
            return generate_openapi(
                router,
                title=title,
                version=version,
                description=description,
                servers=servers,
                namespace=namespace,
            )

    def generate_openapi_with_config_for_test(
        self,
        router: Router,
        config: OpenAPIGeneratorConfig,
        namespace: str = "test",
    ) -> OpenAPI:
        """Generate OpenAPI spec with custom config after registering router."""

        class TestURLConf:
            urlpatterns = [
                path("test/", include((router.urls, namespace))),
            ]

        with override_settings(ROOT_URLCONF=TestURLConf):
            generator = OpenAPIGenerator(config)
            return generator.generate(router, namespace=namespace)

    def validate_spec_for_test(
        self,
        router: Router,
        schema: dict[str, Any],
        namespace: str = "test",
        **kwargs: Any,
    ) -> None:
        """Generate OpenAPI spec and validate against JSON schema."""

        class TestURLConf:
            urlpatterns = [
                path("test/", include((router.urls, namespace))),
            ]

        with override_settings(ROOT_URLCONF=TestURLConf):
            spec = generate_openapi(
                router,
                title=kwargs.get("title", "Test API"),
                version=kwargs.get("version", "1.0.0"),
                description=kwargs.get("description"),
                servers=kwargs.get("servers"),
                namespace=namespace,
            )
            spec_data = json.loads(
                spec.model_dump_json(by_alias=True, exclude_none=True)
            )
            jsonschema.validate(spec_data, schema)



from collections.abc import Callable
from unittest.mock import Mock
import json

from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest, HttpResponse, HttpResponseBase
from django.test import Client, TestCase, override_settings
from django.urls import include, path, reverse
from django.views.decorators.csrf import csrf_exempt

from thelabtyping.api.openapi import OpenAPISchema
from thelabtyping.api.openapi.models import Server
from thelabtyping.api.routing import HttpMethod, Route, Router, URLPatternStr
from thelabtyping.api.status import Status

from ..sampleapp.views import (
    route_users,
    router,
    router_users_list,
)


class RouterTest(TestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(username="joe", password="password")
        self.client = Client()

    def test_reverse_router_urls(self) -> None:
        self.assertEqual(
            reverse("sampleapp:router-users-list"),
            "/api/router/users/",
        )
        self.assertEqual(
            reverse("sampleapp:router-users-detail", args=(42,)),
            "/api/router/users/42/",
        )
        self.assertEqual(
            reverse("sampleapp:alt-router-users-list"),
            "/api/alt-router/users/",
        )
        self.assertEqual(
            reverse("sampleapp:alt-router-users-detail", args=(42,)),
            "/api/alt-router/users/42/",
        )

    def test_router_index_main(self) -> None:
        url = reverse("sampleapp:index")
        self.assertEqual(url, "/api/router/")
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            {
                "sampleapp:index": "http://testserver/api/router/",
                "sampleapp:router-users-list": "http://testserver/api/router/users/",
            },
        )

    def test_router_index_unnamespaced(self) -> None:
        url = reverse("unnamespaced-router-index")
        self.assertEqual(url, "/unnamespaced-router/")
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            {
                "unnamespaced-router-index": "http://testserver/unnamespaced-router/",
                "unnamespaced-router-users-list": "http://testserver/unnamespaced-router/users/",
            },
        )

    def test_router_dispatch_get(self) -> None:
        resp = self.client.get(reverse("sampleapp:router-users-list"))
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            [
                {
                    "username": "joe",
                    "first_name": "",
                    "last_name": "",
                }
            ],
        )

    def test_router_dispatch_post(self) -> None:
        self.client.login(username="joe", password="password")
        resp = self.client.post(
            reverse("sampleapp:router-users-list"),
            content_type="application/json",
            data={
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            {
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )

    def test_router_dispatch_method_not_allowed(self) -> None:
        self.client.login(username="joe", password="password")
        resp = self.client.put(reverse("sampleapp:router-users-list"))
        self.assertEqual(resp.status_code, Status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_register_duplicate_route(self) -> None:
        with self.assertRaises(ImproperlyConfigured):
            router.route("users/", name="router-users-list-duplicate")

    def test_register_duplicate_route_method(self) -> None:
        with self.assertRaises(ImproperlyConfigured):
            route_users.get(router_users_list)

    def test_register_get(self) -> None:
        router = Router()
        route = router.route("test/", name="test")
        self.assertEqual(route.allowed_methods, set())
        route.get(router_users_list)
        self.assertEqual(route.allowed_methods, {HttpMethod.GET})

    def test_register_post(self) -> None:
        router = Router()
        route = router.route("test/", name="test")
        self.assertEqual(route.allowed_methods, set())
        route.post(router_users_list)
        self.assertEqual(route.allowed_methods, {HttpMethod.POST})

    def test_register_put(self) -> None:
        router = Router()
        route = router.route("test/", name="test")
        self.assertEqual(route.allowed_methods, set())
        route.put(router_users_list)
        self.assertEqual(route.allowed_methods, {HttpMethod.PUT})

    def test_register_patch(self) -> None:
        router = Router()
        route = router.route("test/", name="test")
        self.assertEqual(route.allowed_methods, set())
        route.patch(router_users_list)
        self.assertEqual(route.allowed_methods, {HttpMethod.PATCH})

    def test_register_delete(self) -> None:
        router = Router()
        route = router.route("test/", name="test")
        self.assertEqual(route.allowed_methods, set())
        route.delete(router_users_list)
        self.assertEqual(route.allowed_methods, {HttpMethod.DELETE})

    def test_register_during_route_creation(self) -> None:
        router = Router()
        route = router.route(
            "test/",
            name="test",
            get=router_users_list,
            post=router_users_list,
        )
        self.assertEqual(
            route.allowed_methods,
            {
                HttpMethod.GET,
                HttpMethod.POST,
            },
        )


class TestRouteDecorators(TestCase):
    """Tests for Route and Router decorator support."""

    def _dummy_view(
        self, request: HttpRequest, *args: object, **kwargs: object
    ) -> HttpResponse:
        """A simple view function for testing."""
        return HttpResponse("OK")

    def test_route_without_decorators_returns_dispatch(self) -> None:
        """Routes without decorators should return dispatch unchanged."""
        router = Router(enable_index=False)
        route = router.route("test/", name="test", get=self._dummy_view)
        # Without decorators, view should be the same as dispatch
        self.assertEqual(route.view, route.dispatch)

    def test_route_with_csrf_exempt_decorator(self) -> None:
        """csrf_exempt decorator should set attribute on view."""
        router = Router(enable_index=False)
        route = router.route(
            "test/",
            name="test",
            get=self._dummy_view,
            decorators=[csrf_exempt],
        )
        # The view property should have csrf_exempt attribute
        self.assertTrue(hasattr(route.view, "csrf_exempt"))
        self.assertTrue(route.view.csrf_exempt)  # type: ignore[attr-defined]

    def test_multiple_decorators_applied_in_order(self) -> None:
        """Decorators are applied in list order (last is outermost)."""
        call_order: list[str] = []

        def decorator_a(
            fn: Callable[..., HttpResponseBase],
        ) -> Callable[..., HttpResponseBase]:
            def wrapper(
                request: HttpRequest, *args: object, **kwargs: object
            ) -> HttpResponseBase:
                call_order.append("a")
                return fn(request, *args, **kwargs)

            return wrapper

        def decorator_b(
            fn: Callable[..., HttpResponseBase],
        ) -> Callable[..., HttpResponseBase]:
            def wrapper(
                request: HttpRequest, *args: object, **kwargs: object
            ) -> HttpResponseBase:
                call_order.append("b")
                return fn(request, *args, **kwargs)

            return wrapper

        router = Router(enable_index=False)
        route = router.route(
            "test/",
            name="test",
            get=self._dummy_view,
            decorators=[decorator_a, decorator_b],
        )

        # Call the decorated view
        request = Mock(spec=HttpRequest)
        request.method = "GET"
        route.view(request)

        # decorator_b is outermost (applied last), called first
        self.assertEqual(call_order, ["b", "a"])

    def test_view_property_is_cached(self) -> None:
        """Multiple calls to .view return same instance."""
        router = Router(enable_index=False)
        route = router.route(
            "test/",
            name="test",
            get=self._dummy_view,
            decorators=[csrf_exempt],
        )
        view1 = route.view
        view2 = route.view
        self.assertIs(view1, view2)

    def test_router_urls_uses_decorated_view(self) -> None:
        """Router.urls should use route.view, not route.dispatch."""
        router = Router(enable_index=False)
        router.route(
            "test/",
            name="test",
            get=self._dummy_view,
            decorators=[csrf_exempt],
        )
        urls = router.urls
        self.assertEqual(len(urls), 1)
        url_pattern = urls[0]
        # The callback should be the decorated view with csrf_exempt
        self.assertTrue(hasattr(url_pattern.callback, "csrf_exempt"))
        self.assertTrue(url_pattern.callback.csrf_exempt)  # type: ignore[attr-defined]

    def test_router_level_decorators_apply_to_all_routes(self) -> None:
        """Router-level decorators should apply to all routes."""
        router = Router(enable_index=False, decorators=[csrf_exempt])
        route1 = router.route("test1/", name="test1", get=self._dummy_view)
        route2 = router.route("test2/", name="test2", get=self._dummy_view)

        # Both routes should have csrf_exempt applied
        self.assertTrue(hasattr(route1.view, "csrf_exempt"))
        self.assertTrue(route1.view.csrf_exempt)  # type: ignore[attr-defined]
        self.assertTrue(hasattr(route2.view, "csrf_exempt"))
        self.assertTrue(route2.view.csrf_exempt)  # type: ignore[attr-defined]

    def test_router_and_route_decorators_combined(self) -> None:
        """Router decorators apply first (inner), route decorators on top (outer)."""
        call_order: list[str] = []

        def router_decorator(
            fn: Callable[..., HttpResponseBase],
        ) -> Callable[..., HttpResponseBase]:
            def wrapper(
                request: HttpRequest, *args: object, **kwargs: object
            ) -> HttpResponseBase:
                call_order.append("router")
                return fn(request, *args, **kwargs)

            return wrapper

        def route_decorator(
            fn: Callable[..., HttpResponseBase],
        ) -> Callable[..., HttpResponseBase]:
            def wrapper(
                request: HttpRequest, *args: object, **kwargs: object
            ) -> HttpResponseBase:
                call_order.append("route")
                return fn(request, *args, **kwargs)

            return wrapper

        router = Router(enable_index=False, decorators=[router_decorator])
        route = router.route(
            "test/",
            name="test",
            get=self._dummy_view,
            decorators=[route_decorator],
        )

        # Call the decorated view
        request = Mock(spec=HttpRequest)
        request.method = "GET"
        route.view(request)

        # Route decorator is outermost, called first
        self.assertEqual(call_order, ["route", "router"])

    def test_route_created_outside_router_with_decorators(self) -> None:
        """Route can be created directly with decorators."""
        route = Route(name="test", decorators=[csrf_exempt])
        route.get(self._dummy_view)

        self.assertTrue(hasattr(route.view, "csrf_exempt"))
        self.assertTrue(route.view.csrf_exempt)  # type: ignore[attr-defined]


class TestAutoOpenAPI(TestCase):
    """Tests for auto-OpenAPI endpoint feature."""

    def test_openapi_disabled_by_default(self) -> None:
        """No openapi route when no config provided."""
        router = Router(enable_index=False)
        router.route("users/", name="users")

        # Should not have openapi.json route
        route_patterns = [str(pattern) for pattern in router.routes.keys()]
        self.assertNotIn("openapi.json", route_patterns)

    def test_openapi_endpoint_registered(self) -> None:
        """OpenAPI route is registered when schemas is provided."""
        router = Router(
            enable_index=False,
            schemas={
                "openapi.json": OpenAPISchema(
                    title="Test API",
                    version="1.0.0",
                ),
            },
        )
        router.route("users/", name="users")

        # Should have openapi.json route
        route_patterns = [str(pattern) for pattern in router.routes.keys()]
        self.assertIn("openapi.json", route_patterns)

    def test_openapi_custom_url_pattern(self) -> None:
        """Custom URL pattern for OpenAPI endpoint."""
        router = Router(
            enable_index=False,
            schemas={
                "docs/openapi.json": OpenAPISchema(
                    title="Test API",
                    version="1.0.0",
                ),
            },
        )

        # Should have custom route
        route_patterns = [str(pattern) for pattern in router.routes.keys()]
        self.assertIn("docs/openapi.json", route_patterns)
        self.assertNotIn("openapi.json", route_patterns)

    def test_openapi_serves_valid_spec(self) -> None:
        """OpenAPI endpoint returns valid OpenAPI JSON."""
        router = Router(
            basename="test",
            enable_index=False,
            schemas={
                "openapi.json": OpenAPISchema(
                    title="Test API",
                    version="2.0.0",
                    description="A test API",
                    cache_timeout=None,  # Disable caching for test with mock request
                    require_auth=False,  # Disable auth for unit test
                ),
            },
        )
        router.route("users/", name="users")

        # Get the openapi route
        openapi_route = router.routes.get(URLPatternStr("openapi.json"))
        self.assertIsNotNone(openapi_route)
        assert openapi_route is not None  # for mypy

        # Create a mock request with resolver_match for namespace
        request = Mock(spec=HttpRequest)
        request.method = "GET"
        request.build_absolute_uri = Mock(return_value="http://testserver/")
        request.resolver_match = Mock()
        request.resolver_match.namespace = "test"

        # Set up URL config for reverse() to work
        class TestURLConf:
            urlpatterns = [
                path("test/", include((router.urls, "test"))),
            ]

        # Call the view with URL config in place
        with override_settings(ROOT_URLCONF=TestURLConf):
            response = openapi_route.view(request)

        # Check response
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["Content-Type"], "application/json")

        # Parse and validate the JSON
        assert hasattr(response, "content")  # APIResponse has content
        data = json.loads(response.content)
        self.assertEqual(data["openapi"], "3.1.0")
        self.assertEqual(data["info"]["title"], "Test API")
        self.assertEqual(data["info"]["version"], "2.0.0")
        self.assertEqual(data["info"]["description"], "A test API")
        self.assertIn("/test/users/", data["paths"])
        self.assertIn("/test/openapi.json", data["paths"])

    def test_openapi_requires_auth_by_default(self) -> None:
        """OpenAPI endpoint requires authentication by default."""
        router = Router(
            enable_index=False,
            schemas={
                "openapi.json": OpenAPISchema(
                    title="Test API",
                    version="1.0.0",
                    cache_timeout=None,  # Disable caching for test
                ),
            },
        )

        # Get the openapi route
        openapi_route = router.routes.get(URLPatternStr("openapi.json"))
        self.assertIsNotNone(openapi_route)
        assert openapi_route is not None

        # Create a mock request with anonymous user
        request = Mock(spec=HttpRequest)
        request.method = "GET"
        request.user = Mock()
        request.user.is_anonymous = True

        # Call the view - should return 403
        response = openapi_route.view(request)
        self.assertEqual(response.status_code, Status.HTTP_403_FORBIDDEN)

    def test_openapi_allows_authenticated_users(self) -> None:
        """OpenAPI endpoint allows authenticated users when auth is required."""
        router = Router(
            enable_index=False,
            schemas={
                "openapi.json": OpenAPISchema(
                    title="Test API",
                    version="1.0.0",
                    cache_timeout=None,  # Disable caching for test
                ),
            },
        )

        # Get the openapi route
        openapi_route = router.routes.get(URLPatternStr("openapi.json"))
        self.assertIsNotNone(openapi_route)
        assert openapi_route is not None

        # Create a mock request with authenticated user
        request = Mock(spec=HttpRequest)
        request.method = "GET"
        request.user = Mock()
        request.user.is_anonymous = False
        request.build_absolute_uri = Mock(return_value="http://testserver/")
        request.resolver_match = Mock()
        request.resolver_match.namespace = "test"

        # Set up URL config for reverse() to work
        class TestURLConf:
            urlpatterns = [
                path("test/", include((router.urls, "test"))),
            ]

        # Call the view - should return 200
        with override_settings(ROOT_URLCONF=TestURLConf):
            response = openapi_route.view(request)
        self.assertEqual(response.status_code, Status.HTTP_200_OK)

        # Verify it's valid JSON
        self.assertIsInstance(response, HttpResponse)
        assert isinstance(response, HttpResponse)
        data = json.loads(response.content)
        self.assertEqual(data["openapi"], "3.1.0")

    def test_openapi_public_when_auth_disabled(self) -> None:
        """OpenAPI endpoint is public when require_auth is False."""
        router = Router(
            enable_index=False,
            schemas={
                "openapi.json": OpenAPISchema(
                    title="Test API",
                    version="1.0.0",
                    cache_timeout=None,
                    require_auth=False,  # Allow public access
                ),
            },
        )

        # Get the openapi route
        openapi_route = router.routes.get(URLPatternStr("openapi.json"))
        self.assertIsNotNone(openapi_route)
        assert openapi_route is not None

        # Create a mock request with anonymous user
        request = Mock(spec=HttpRequest)
        request.method = "GET"
        request.user = Mock()
        request.user.is_anonymous = True
        request.build_absolute_uri = Mock(return_value="http://testserver/")
        request.resolver_match = Mock()
        request.resolver_match.namespace = "test"

        # Set up URL config for reverse() to work
        class TestURLConf:
            urlpatterns = [
                path("test/", include((router.urls, "test"))),
            ]

        # Call the view - should return 200 even for anonymous user
        with override_settings(ROOT_URLCONF=TestURLConf):
            response = openapi_route.view(request)
        self.assertEqual(response.status_code, Status.HTTP_200_OK)

        # Verify it's valid JSON
        self.assertIsInstance(response, HttpResponse)
        assert isinstance(response, HttpResponse)
        data = json.loads(response.content)
        self.assertEqual(data["openapi"], "3.1.0")

    def test_openapi_caching_enabled_by_default(self) -> None:
        """OpenAPI schema generator has caching enabled by default (3600 seconds)."""
        schema = OpenAPISchema(
            title="Test API",
            version="1.0.0",
        )

        # Default cache_timeout is 3600
        self.assertEqual(schema.cache_timeout, 3600)

    def test_openapi_caching_disabled_when_none(self) -> None:
        """OpenAPI schema generator can have caching disabled."""
        schema = OpenAPISchema(
            title="Test API",
            version="1.0.0",
            cache_timeout=None,
        )

        # cache_timeout is None when disabled
        self.assertIsNone(schema.cache_timeout)

    def test_openapi_custom_cache_timeout(self) -> None:
        """OpenAPI schema generator respects custom cache_timeout value."""
        schema = OpenAPISchema(
            title="Test API",
            version="1.0.0",
            cache_timeout=7200,  # 2 hours
        )

        # Custom cache_timeout is preserved
        self.assertEqual(schema.cache_timeout, 7200)

    def test_openapi_auto_populates_servers_from_request(self) -> None:
        """OpenAPI spec auto-populates servers from request when not provided."""
        router = Router(
            enable_index=False,
            schemas={
                "openapi.json": OpenAPISchema(
                    title="Test API",
                    version="1.0.0",
                    cache_timeout=None,
                    require_auth=False,
                    # No servers provided - should auto-populate
                ),
            },
        )

        openapi_route = router.routes.get(URLPatternStr("openapi.json"))
        assert openapi_route is not None

        request = Mock(spec=HttpRequest)
        request.method = "GET"
        request.build_absolute_uri = Mock(return_value="https://api.example.com/")
        request.resolver_match = Mock()
        request.resolver_match.namespace = "test"

        # Set up URL config for reverse() to work
        class TestURLConf:
            urlpatterns = [
                path("test/", include((router.urls, "test"))),
            ]

        with override_settings(ROOT_URLCONF=TestURLConf):
            response = openapi_route.view(request)
        self.assertEqual(response.status_code, Status.HTTP_200_OK)

        self.assertIsInstance(response, HttpResponse)
        assert isinstance(response, HttpResponse)
        data = json.loads(response.content)

        # Server should be auto-populated from request
        self.assertIn("servers", data)
        self.assertEqual(len(data["servers"]), 1)
        self.assertEqual(data["servers"][0]["url"], "https://api.example.com")

    def test_openapi_uses_explicit_servers_when_provided(self) -> None:
        """OpenAPI spec uses explicit servers when provided."""
        router = Router(
            enable_index=False,
            schemas={
                "openapi.json": OpenAPISchema(
                    title="Test API",
                    version="1.0.0",
                    cache_timeout=None,
                    require_auth=False,
                    servers=[
                        Server(url="https://prod.example.com"),
                        Server(url="https://staging.example.com"),
                    ],
                ),
            },
        )

        openapi_route = router.routes.get(URLPatternStr("openapi.json"))
        assert openapi_route is not None

        request = Mock(spec=HttpRequest)
        request.method = "GET"
        # This should be ignored when servers are explicitly provided
        request.build_absolute_uri = Mock(return_value="https://ignored.example.com/")
        request.resolver_match = Mock()
        request.resolver_match.namespace = "test"

        # Set up URL config for reverse() to work
        class TestURLConf:
            urlpatterns = [
                path("test/", include((router.urls, "test"))),
            ]

        with override_settings(ROOT_URLCONF=TestURLConf):
            response = openapi_route.view(request)
        self.assertEqual(response.status_code, Status.HTTP_200_OK)

        self.assertIsInstance(response, HttpResponse)
        assert isinstance(response, HttpResponse)
        data = json.loads(response.content)

        # Explicit servers should be used, not auto-populated
        self.assertIn("servers", data)
        self.assertEqual(len(data["servers"]), 2)
        self.assertEqual(data["servers"][0]["url"], "https://prod.example.com")
        self.assertEqual(data["servers"][1]["url"], "https://staging.example.com")



from typing import Literal, assert_type

from django.contrib.auth.models import AnonymousUser, User
from django.shortcuts import get_object_or_404
from django.utils.decorators import method_decorator
from django.views.generic import View

from thelabtyping.abc import ListOf
from thelabtyping.api.requests import (
    AuthdEmptyTypedRequest,
    AuthdTypedRequest,
    AuthdTypedRequestBody,
    AuthdTypedRequestQuery,
    EmptyTypedRequest,
    TypedRequest,
    TypedRequestBody,
    TypedRequestQuery,
)
from thelabtyping.api.responses import APIResponse
from thelabtyping.api.routing import Router
from thelabtyping.api.serializers import Empty
from thelabtyping.api.views import validate

from .serializers import APIUser, UserSearchQuery


class UserView(View):
    @method_decorator(validate())
    def get(
        self, request: TypedRequestQuery[UserSearchQuery]
    ) -> APIResponse[ListOf[APIUser]]:
        # Some assertions to prove the `validate` decorator is doing what it's
        # supposed to
        assert_type(request, TypedRequest[UserSearchQuery, Empty])
        assert_type(request.user, User | AnonymousUser)
        assert_type(request.validated_querystring, UserSearchQuery)
        assert_type(request.validated_body, Empty)
        # Do the same checks at runtime
        assert isinstance(request, TypedRequest)
        assert isinstance(request.user, (User | AnonymousUser))
        assert isinstance(request.validated_querystring, UserSearchQuery)
        assert isinstance(request.validated_body, Empty)
        # Use the UserSearchQuery model to get he queryset of users to return
        users, errs = APIUser.list_from_django(
            request,
            request.validated_querystring.get_queryset(),
        )
        return APIResponse(users)

    @method_decorator(validate())
    def post(self, request: AuthdTypedRequestBody[APIUser]) -> APIResponse[APIUser]:
        # Some assertions to prove the `validate` decorator is doing what it's
        # supposed to
        assert_type(request, TypedRequest[Empty, APIUser, User])
        assert_type(request.user, User)
        assert_type(request.validated_querystring, Empty)
        assert_type(request.validated_body, APIUser)
        # Do the same checks at runtime
        assert isinstance(request, TypedRequest)
        assert isinstance(request.user, User)
        assert isinstance(request.validated_querystring, Empty)
        assert isinstance(request.validated_body, APIUser)
        # Save the user model and return it
        result = (
            # Take the input data
            request.validated_body
            # Create the Django user
            .create(request)
            # Then serialize it back into an APIUser
            .and_then(lambda djuser: APIUser.from_django(request, djuser))
        )
        return APIResponse(result)


router = Router()
route_users = router.route("users/", name="router-users-list")
route_user_detail = router.route("users/<int:pk>/", name="router-users-detail")


@route_users.get
@validate()
def router_users_list(
    request: TypedRequestQuery[UserSearchQuery],
) -> APIResponse[ListOf[APIUser]]:
    # Use the UserSearchQuery model to get he queryset of users to return
    users, errs = APIUser.list_from_django(
        request,
        request.validated_querystring.get_queryset(),
    )
    return APIResponse(users)


@route_user_detail.get
@validate()
def router_users_detail(
    request: EmptyTypedRequest,
    pk: int,
) -> APIResponse[APIUser]:
    # Use the UserSearchQuery model to get he queryset of users to return
    user = get_object_or_404(User, pk=pk)
    return APIResponse(APIUser.from_django(request, user))


@route_users.post
@validate()
def router_users_post(request: AuthdTypedRequestBody[APIUser]) -> APIResponse[APIUser]:
    # Save the user model and return it
    result = (
        # Take the input data
        request.validated_body
        # Create the Django user
        .create(request)
        # Then serialize it back into an APIUser
        .and_then(lambda djuser: APIUser.from_django(request, djuser))
    )
    return APIResponse(result)


@validate()
def get_test_typed_request_body(
    request: TypedRequestBody[APIUser],
) -> APIResponse[Empty]:
    # Some assertions to prove the `validate` decorator is doing what it's
    # supposed to
    assert_type(request, TypedRequest[Empty, APIUser])
    assert_type(request.user, User | AnonymousUser)
    assert_type(request.validated_querystring, Empty)
    assert_type(request.validated_body, APIUser)
    return APIResponse(Empty())


@validate()
def get_test_empty_request(request: EmptyTypedRequest) -> APIResponse[Empty]:
    # Some assertions to prove the `validate` decorator is doing what it's
    # supposed to
    assert_type(request, TypedRequest[Empty, Empty])
    assert_type(request.user, User | AnonymousUser)
    assert_type(request.validated_querystring, Empty)
    assert_type(request.validated_body, Empty)
    return APIResponse(Empty())


@validate()
def get_test_authd_typed_request(
    request: AuthdTypedRequest[UserSearchQuery, APIUser],
) -> APIResponse[Empty]:
    # Some assertions to prove the `validate` decorator is doing what it's
    # supposed to
    assert_type(request, TypedRequest[UserSearchQuery, APIUser, User])
    assert_type(request.user, User)
    assert_type(request.user.is_authenticated, Literal[True])
    assert_type(request.user.is_anonymous, Literal[False])
    assert_type(request.validated_querystring, UserSearchQuery)
    assert_type(request.validated_body, APIUser)
    return APIResponse(Empty())


@validate()
def get_test_authd_typed_request_query(
    request: AuthdTypedRequestQuery[UserSearchQuery],
) -> APIResponse[Empty]:
    # Some assertions to prove the `validate` decorator is doing what it's
    # supposed to
    assert_type(request, TypedRequest[UserSearchQuery, Empty, User])
    assert_type(request.user, User)
    assert_type(request.user.is_authenticated, Literal[True])
    assert_type(request.user.is_anonymous, Literal[False])
    assert_type(request.validated_querystring, UserSearchQuery)
    assert_type(request.validated_body, Empty)
    return APIResponse(Empty())


@validate()
def get_test_authd_typed_request_body(
    request: AuthdTypedRequestBody[APIUser],
) -> APIResponse[Empty]:
    # Some assertions to prove the `validate` decorator is doing what it's
    # supposed to
    assert_type(request, TypedRequest[Empty, APIUser, User])
    assert_type(request.user, User)
    assert_type(request.user.is_authenticated, Literal[True])
    assert_type(request.user.is_anonymous, Literal[False])
    assert_type(request.validated_querystring, Empty)
    assert_type(request.validated_body, APIUser)
    return APIResponse(Empty())


@validate()
def get_test_authd_typed_request_empty(
    request: AuthdEmptyTypedRequest,
) -> APIResponse[Empty]:
    # Some assertions to prove the `validate` decorator is doing what it's
    # supposed to
    assert_type(request, TypedRequest[Empty, Empty, User])
    assert_type(request.user, User)
    assert_type(request.user.is_authenticated, Literal[True])
    assert_type(request.user.is_anonymous, Literal[False])
    assert_type(request.validated_querystring, Empty)
    assert_type(request.validated_body, Empty)
    return APIResponse(Empty())



from django.test import TestCase

from thelabtyping.api.status import Status


class HttpStatusTest(TestCase):
    def test_statuses(self) -> None:
        self.assertEqual(Status.HTTP_200_OK.is_informational, False)
        self.assertEqual(Status.HTTP_200_OK.is_success, True)
        self.assertEqual(Status.HTTP_200_OK.is_redirect, False)
        self.assertEqual(Status.HTTP_200_OK.is_client_error, False)
        self.assertEqual(Status.HTTP_200_OK.is_server_error, False)

        self.assertEqual(Status.HTTP_403_FORBIDDEN.is_informational, False)
        self.assertEqual(Status.HTTP_403_FORBIDDEN.is_success, False)
        self.assertEqual(Status.HTTP_403_FORBIDDEN.is_redirect, False)
        self.assertEqual(Status.HTTP_403_FORBIDDEN.is_client_error, True)
        self.assertEqual(Status.HTTP_403_FORBIDDEN.is_server_error, False)



"""OpenAPI 3.1 specification Pydantic models."""

from typing import Any, Literal

import pydantic


class Contact(pydantic.BaseModel):
    """OpenAPI Contact Object.

    https://spec.openapis.org/oas/v3.1.0#contact-object
    """

    name: str | None = None
    url: pydantic.HttpUrl | None = None
    email: str | None = None


class License(pydantic.BaseModel):
    """OpenAPI License Object.

    https://spec.openapis.org/oas/v3.1.0#license-object
    """

    name: str
    identifier: str | None = None
    url: pydantic.HttpUrl | None = None


class Info(pydantic.BaseModel):
    """OpenAPI Info Object.

    https://spec.openapis.org/oas/v3.1.0#info-object
    """

    title: str
    version: str
    summary: str | None = None
    description: str | None = None
    termsOfService: pydantic.HttpUrl | None = None
    contact: Contact | None = None
    license: License | None = None


class Server(pydantic.BaseModel):
    """OpenAPI Server Object.

    https://spec.openapis.org/oas/v3.1.0#server-object
    """

    url: str
    description: str | None = None


class ExternalDocumentation(pydantic.BaseModel):
    """OpenAPI External Documentation Object.

    https://spec.openapis.org/oas/v3.1.0#external-documentation-object
    """

    url: pydantic.HttpUrl
    description: str | None = None


class Reference(pydantic.BaseModel):
    """OpenAPI Reference Object.

    https://spec.openapis.org/oas/v3.1.0#reference-object
    """

    ref: str = pydantic.Field(serialization_alias="$ref")

    model_config = pydantic.ConfigDict(
        validate_by_name=True,
        validate_by_alias=False,
        serialize_by_alias=True,
    )


class MediaType(pydantic.BaseModel):
    """OpenAPI Media Type Object.

    https://spec.openapis.org/oas/v3.1.0#media-type-object
    """

    media_type_schema: dict[str, Any] | Reference | None = pydantic.Field(
        default=None, serialization_alias="schema"
    )

    model_config = pydantic.ConfigDict(
        validate_by_name=True,
        validate_by_alias=False,
        serialize_by_alias=True,
    )


class RequestBody(pydantic.BaseModel):
    """OpenAPI Request Body Object.

    https://spec.openapis.org/oas/v3.1.0#request-body-object
    """

    description: str | None = None
    content: dict[str, MediaType]
    required: bool = False


class Parameter(pydantic.BaseModel):
    """OpenAPI Parameter Object.

    https://spec.openapis.org/oas/v3.1.0#parameter-object
    """

    name: str
    param_in: Literal["query", "header", "path", "cookie"] = pydantic.Field(
        serialization_alias="in"
    )
    description: str | None = None
    required: bool = False
    deprecated: bool = False
    param_schema: dict[str, Any] | Reference | None = pydantic.Field(
        default=None, serialization_alias="schema"
    )

    model_config = pydantic.ConfigDict(
        validate_by_name=True,
        validate_by_alias=False,
        serialize_by_alias=True,
    )


class Response(pydantic.BaseModel):
    """OpenAPI Response Object.

    https://spec.openapis.org/oas/v3.1.0#response-object
    """

    description: str
    content: dict[str, MediaType] | None = None


class Operation(pydantic.BaseModel):
    """OpenAPI Operation Object.

    https://spec.openapis.org/oas/v3.1.0#operation-object
    """

    tags: list[str] | None = None
    summary: str | None = None
    description: str | None = None
    operationId: str | None = None
    parameters: list[Parameter | Reference] | None = None
    requestBody: RequestBody | Reference | None = None
    responses: dict[str, Response | Reference]
    deprecated: bool = False
    security: list[dict[str, list[str]]] | None = None


class PathItem(pydantic.BaseModel):
    """OpenAPI Path Item Object.

    https://spec.openapis.org/oas/v3.1.0#path-item-object
    """

    summary: str | None = None
    description: str | None = None
    get: Operation | None = None
    put: Operation | None = None
    post: Operation | None = None
    delete: Operation | None = None
    patch: Operation | None = None
    parameters: list[Parameter | Reference] | None = None


class SecurityScheme(pydantic.BaseModel):
    """OpenAPI Security Scheme Object.

    https://spec.openapis.org/oas/v3.1.0#security-scheme-object
    """

    type: Literal["apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"]
    description: str | None = None
    name: str | None = None
    security_scheme_in: Literal["query", "header", "cookie"] | None = pydantic.Field(
        default=None, serialization_alias="in"
    )
    scheme: str | None = None
    bearerFormat: str | None = None

    model_config = pydantic.ConfigDict(
        validate_by_name=True,
        validate_by_alias=False,
        serialize_by_alias=True,
    )


class Components(pydantic.BaseModel):
    """OpenAPI Components Object.

    https://spec.openapis.org/oas/v3.1.0#components-object
    """

    schemas: dict[str, dict[str, Any]] | None = None
    securitySchemes: dict[str, SecurityScheme] | None = None


class Tag(pydantic.BaseModel):
    """OpenAPI Tag Object.

    https://spec.openapis.org/oas/v3.1.0#tag-object
    """

    name: str
    description: str | None = None
    externalDocs: ExternalDocumentation | None = None


class OpenAPI(pydantic.BaseModel):
    """OpenAPI 3.1 Document Root.

    https://spec.openapis.org/oas/v3.1.0#openapi-object
    """

    openapi: Literal["3.1.0"] = "3.1.0"
    info: Info
    servers: list[Server] | None = None
    paths: dict[str, PathItem]
    components: Components | None = None
    security: list[dict[str, list[str]]] | None = None
    tags: list[Tag] | None = None
    externalDocs: ExternalDocumentation | None = None



"""Tests for type introspection utilities."""

from unittest import TestCase

from django.http import HttpResponse

from thelabtyping.abc import ListOf
from thelabtyping.api.openapi.introspection import (
    infer_view_types,
    is_empty_model,
)
from thelabtyping.api.requests import (
    AuthdEmptyTypedRequest,
    AuthdTypedRequest,
    AuthdTypedRequestBody,
    AuthdTypedRequestQuery,
    EmptyTypedRequest,
    TypedRequest,
    TypedRequestBody,
    TypedRequestQuery,
)
from thelabtyping.api.responses import APIResponse
from thelabtyping.api.serializers import Empty

from .conftest import BodyModel, QueryModel, ResponseModel


class TestIsEmptyModel(TestCase):
    def test_empty_is_empty(self) -> None:
        """Empty model is identified as empty."""
        self.assertTrue(is_empty_model(Empty))

    def test_none_is_empty(self) -> None:
        """None is considered empty."""
        self.assertTrue(is_empty_model(None))

    def test_other_model_not_empty(self) -> None:
        """Other Pydantic models are not empty."""
        self.assertFalse(is_empty_model(QueryModel))
        self.assertFalse(is_empty_model(BodyModel))


class TestInferViewTypesEmptyRequest(TestCase):
    def test_empty_typed_request(self) -> None:
        """EmptyTypedRequest has no query or body models."""

        def view(request: EmptyTypedRequest) -> APIResponse[Empty]:
            return APIResponse(Empty())

        info = infer_view_types(view)
        self.assertIsNone(info.querystring_model)
        self.assertIsNone(info.body_model)
        self.assertFalse(info.requires_auth)

    def test_authd_empty_typed_request(self) -> None:
        """AuthdEmptyTypedRequest requires auth but has no models."""

        def view(request: AuthdEmptyTypedRequest) -> APIResponse[Empty]:
            return APIResponse(Empty())

        info = infer_view_types(view)
        self.assertIsNone(info.querystring_model)
        self.assertIsNone(info.body_model)
        self.assertTrue(info.requires_auth)


class TestInferViewTypesQueryRequest(TestCase):
    def test_typed_request_query(self) -> None:
        """TypedRequestQuery extracts query model."""

        def view(request: TypedRequestQuery[QueryModel]) -> APIResponse[Empty]:
            return APIResponse(Empty())

        info = infer_view_types(view)
        self.assertEqual(info.querystring_model, QueryModel)
        self.assertIsNone(info.body_model)
        self.assertFalse(info.requires_auth)

    def test_authd_typed_request_query(self) -> None:
        """AuthdTypedRequestQuery extracts query model and requires auth."""

        def view(request: AuthdTypedRequestQuery[QueryModel]) -> APIResponse[Empty]:
            return APIResponse(Empty())

        info = infer_view_types(view)
        self.assertEqual(info.querystring_model, QueryModel)
        self.assertIsNone(info.body_model)
        self.assertTrue(info.requires_auth)


class TestInferViewTypesBodyRequest(TestCase):
    def test_typed_request_body(self) -> None:
        """TypedRequestBody extracts body model."""

        def view(request: TypedRequestBody[BodyModel]) -> APIResponse[Empty]:
            return APIResponse(Empty())

        info = infer_view_types(view)
        self.assertIsNone(info.querystring_model)
        self.assertEqual(info.body_model, BodyModel)
        self.assertFalse(info.requires_auth)

    def test_authd_typed_request_body(self) -> None:
        """AuthdTypedRequestBody extracts body model and requires auth."""

        def view(request: AuthdTypedRequestBody[BodyModel]) -> APIResponse[Empty]:
            return APIResponse(Empty())

        info = infer_view_types(view)
        self.assertIsNone(info.querystring_model)
        self.assertEqual(info.body_model, BodyModel)
        self.assertTrue(info.requires_auth)


class TestInferViewTypesFullRequest(TestCase):
    def test_typed_request(self) -> None:
        """TypedRequest extracts both query and body models."""

        def view(
            request: TypedRequest[QueryModel, BodyModel],
        ) -> APIResponse[ResponseModel]:
            return APIResponse(ResponseModel(id=1, name="test"))

        info = infer_view_types(view)
        self.assertEqual(info.querystring_model, QueryModel)
        self.assertEqual(info.body_model, BodyModel)
        self.assertFalse(info.requires_auth)

    def test_authd_typed_request(self) -> None:
        """AuthdTypedRequest extracts both models and requires auth."""

        def view(
            request: AuthdTypedRequest[QueryModel, BodyModel],
        ) -> APIResponse[ResponseModel]:
            return APIResponse(ResponseModel(id=1, name="test"))

        info = infer_view_types(view)
        self.assertEqual(info.querystring_model, QueryModel)
        self.assertEqual(info.body_model, BodyModel)
        self.assertTrue(info.requires_auth)


class TestInferViewTypesResponse(TestCase):
    def test_api_response_model(self) -> None:
        """APIResponse[T] extracts response model."""

        def view(request: EmptyTypedRequest) -> APIResponse[ResponseModel]:
            return APIResponse(ResponseModel(id=1, name="test"))

        info = infer_view_types(view)
        self.assertEqual(info.response_model, ResponseModel)

    def test_api_response_list_of(self) -> None:
        """APIResponse[ListOf[T]] extracts ListOf response model."""

        def view(request: EmptyTypedRequest) -> APIResponse[ListOf[ResponseModel]]:
            return APIResponse(ListOf[ResponseModel]([]))

        info = infer_view_types(view)
        # ListOf is a generic alias, but we should get the actual type
        self.assertIsNotNone(info.response_model)

    def test_api_response_empty(self) -> None:
        """APIResponse[Empty] has no response model."""

        def view(request: EmptyTypedRequest) -> APIResponse[Empty]:
            return APIResponse(Empty())

        info = infer_view_types(view)
        # Empty is the response model, but we treat it as None
        self.assertIsNone(info.response_model)


class TestInferViewTypesEdgeCases(TestCase):
    def test_function_with_path_params(self) -> None:
        """Function with path parameters still works."""

        def view(
            request: TypedRequestQuery[QueryModel], pk: int
        ) -> APIResponse[ResponseModel]:
            return APIResponse(ResponseModel(id=pk, name="test"))

        info = infer_view_types(view)
        self.assertEqual(info.querystring_model, QueryModel)
        self.assertEqual(info.response_model, ResponseModel)

    def test_unannotated_request(self) -> None:
        """Function without request type hint returns empty info."""

        def view(request) -> HttpResponse:  # type: ignore[no-untyped-def]
            return HttpResponse()

        info = infer_view_types(view)
        self.assertIsNone(info.querystring_model)
        self.assertIsNone(info.body_model)
        self.assertFalse(info.requires_auth)

    def test_non_api_response_return(self) -> None:
        """Function with non-APIResponse return has no response model."""

        def view(request: EmptyTypedRequest) -> HttpResponse:
            return HttpResponse()

        info = infer_view_types(view)
        self.assertIsNone(info.response_model)



import uuid

from django.contrib.auth.models import User
from django.test import RequestFactory, TestCase
import pydantic_core

from thelabtyping.abc import ListOf

from ..sampleapp.serializers import APIUser, APIUserWithCustomList


class APIObjTest(TestCase):
    def setUp(self) -> None:
        self.rfact = RequestFactory()
        self.request = self.rfact.get("/api/users/")
        self.users = [
            User.objects.create_user(
                username=str(uuid.uuid4()),
                first_name=f"joe-{i}",
                last_name=f"schmoe-{i}",
            )
            for i in range(3)
        ]

    def test_serialize_single_object(self) -> None:
        maybeobj = APIUser.from_django(self.request, self.users[0])
        assert maybeobj.is_ok
        obj = maybeobj.ok()
        self.assertIsInstance(obj, APIUser)
        self.assertEqual(
            obj.model_dump(),
            {
                "username": self.users[0].username,
                "first_name": "joe-0",
                "last_name": "schmoe-0",
            },
        )

    def test_serialize_object_list(self) -> None:
        obj_list, errs = APIUser.list_from_django(self.request, self.users)
        self.assertIsInstance(obj_list, ListOf)
        self.assertIsInstance(obj_list[0], APIUser)
        self.assertEqual(len(obj_list), len(self.users))
        self.assertEqual(len(errs), 0)
        self.assertEqual(
            obj_list.model_dump(),
            [
                {
                    "username": self.users[0].username,
                    "first_name": "joe-0",
                    "last_name": "schmoe-0",
                },
                {
                    "username": self.users[1].username,
                    "first_name": "joe-1",
                    "last_name": "schmoe-1",
                },
                {
                    "username": self.users[2].username,
                    "first_name": "joe-2",
                    "last_name": "schmoe-2",
                },
            ],
        )

    def test_serialize_object_list_custom_list_class(self) -> None:
        obj_list, errs = APIUserWithCustomList.list_from_django(
            self.request,
            self.users,
        )
        self.assertIsInstance(obj_list, APIUserWithCustomList.List)
        self.assertIsInstance(obj_list[0], APIUserWithCustomList)
        self.assertEqual(len(obj_list), len(self.users))
        self.assertEqual(len(errs), 0)
        self.assertEqual(
            obj_list.model_dump(),
            [
                {
                    "username": self.users[0].username,
                    "first_name": "joe-0",
                    "last_name": "schmoe-0",
                },
                {
                    "username": self.users[1].username,
                    "first_name": "joe-1",
                    "last_name": "schmoe-1",
                },
                {
                    "username": self.users[2].username,
                    "first_name": "joe-2",
                    "last_name": "schmoe-2",
                },
            ],
        )

    def test_catches_validation_errs(self) -> None:
        # Force a validation error by setting a first_name to an integer
        self.users[1].first_name = 42
        # Result list should omit the invalid object and include it as an error
        obj_list, errs = APIUser.list_from_django(self.request, self.users)
        self.assertIsInstance(obj_list, ListOf)
        self.assertEqual(len(obj_list), len(self.users) - 1)
        self.assertEqual(len(errs), 1)
        self.assertEqual(
            obj_list.model_dump(),
            [
                {
                    "username": self.users[0].username,
                    "first_name": "joe-0",
                    "last_name": "schmoe-0",
                },
                {
                    "username": self.users[2].username,
                    "first_name": "joe-2",
                    "last_name": "schmoe-2",
                },
            ],
        )
        self.assertIsInstance(errs[0], pydantic_core.ValidationError)
        self.assertTrue(
            str(errs[0]).startswith(
                "1 validation error for APIUser\n"
                "first_name\n"
                "  Input should be a valid string [type=string_type, input_value=42, input_type=int]"
            ),
        )



from typing import Literal, assert_type
from unittest import TestCase

from thelabtyping.guards import is_literal_factory


class IsLiteralTest(TestCase):
    def test_normal_literal(self) -> None:
        MyLiteral = Literal["foo", "bar"]
        is_literal = is_literal_factory(MyLiteral)

        scenarios: list[tuple[str, bool]] = [
            ("a", False),
            ("foo", True),
            ("bar", True),
            ("baz", False),
        ]
        for arg, should_match in scenarios:
            with self.subTest(
                f"{arg} {'is literal' if should_match else 'is not literal'}"
            ):
                # Outside the type guard, some_str is just a string
                assert_type(arg, str)
                if is_literal(arg):
                    # Assert that mypy correctly narrowed the str type into the
                    # literal type
                    assert_type(arg, MyLiteral)
                    did_match = True
                else:
                    did_match = False
                # Check if matching/not-matching was expected.
                self.assertEqual(did_match, should_match)

    def test_alias_literal(self) -> None:
        type MyLiteral = Literal["foo", "bar"]
        is_literal = is_literal_factory(MyLiteral)

        scenarios: list[tuple[str, bool]] = [
            ("a", False),
            ("foo", True),
            ("bar", True),
            ("baz", False),
        ]
        for arg, should_match in scenarios:
            with self.subTest(
                f"{arg} {'is literal' if should_match else 'is not literal'}"
            ):
                # Outside the type guard, some_str is just a string
                assert_type(arg, str)
                if is_literal(arg):
                    # Assert that mypy correctly narrowed the str type into the
                    # literal type
                    assert_type(arg, MyLiteral)
                    did_match = True
                else:
                    did_match = False
                # Check if matching/not-matching was expected.
                self.assertEqual(did_match, should_match)



from collections.abc import Callable
from typing import _LiteralGenericAlias  # type:ignore[attr-defined]
from typing import (
    TypeAliasType,
    TypeIs,
    get_args,
)


def is_literal_factory[T](lit: T) -> Callable[[object], TypeIs[T]]:
    """
    Given a `Literal` type, return a TypeIs type guard to determine if a
    variable is a member of that Literal. Example usage:

    >>> from typing import Literal
    >>> type MyLiteral = Literal["foo", "bar"]
    >>> is_literal = is_literal_factory(MyLiteral)
    >>> is_literal("foo")
    True
    >>> is_literal("FOO")
    False
    """

    def is_literal(arg: object) -> TypeIs[T]:
        """
        Type guard which narrows a type down to being a member of a Literal
        """
        # If the literal was defined with the `type` keyword, then we ned to
        # force it to resolve by getting lit.__value__.
        resolved_lit = lit.__value__ if (type(lit) is TypeAliasType) else lit
        assert isinstance(resolved_lit, _LiteralGenericAlias)
        return arg in get_args(resolved_lit)

    return is_literal



from unittest import TestCase

import pydantic

from thelabtyping.abc import DictOf, ListOf


class Person(pydantic.BaseModel):
    name: str


class PersonList(ListOf[Person]):
    pass


Data = DictOf[str, int]


class ListOfTest(TestCase):
    def setUp(self) -> None:
        self.people = PersonList.model_validate(
            [
                {"name": "Jack"},
                {"name": "Diane"},
            ]
        )

    def test_isinstance(self) -> None:
        self.assertIsInstance(self.people, PersonList)
        self.assertIsInstance(self.people[0], Person)

    def test_index_accessors(self) -> None:
        self.assertEqual(self.people[0].name, "Jack")
        self.assertEqual(self.people[1].name, "Diane")

    def test_iteration(self) -> None:
        names = ["Jack", "Diane"]
        for i, person in enumerate(self.people):
            self.assertEqual(person.name, names[i])


class DictOfTest(TestCase):
    def test_dunders(self) -> None:
        data = Data({})
        self.assertEqual(data.model_dump(), {})

        # __setitem__
        data["test"] = 42
        self.assertEqual(data.model_dump(), {"test": 42})

        # __getitem__
        self.assertEqual(data["test"], 42)

        # __contains__
        self.assertIn("test", data)
        self.assertNotIn("foo", data)

        # __iter__
        self.assertEqual(list(data), ["test"])

        # __len__
        self.assertEqual(len(data), 1)

        # keys
        self.assertEqual(list(data.keys()), ["test"])

        # values
        self.assertEqual(list(data.values()), [42])

        # items
        self.assertEqual(list(data.items()), [("test", 42)])

        # get
        self.assertEqual(data.get("test"), 42)
        self.assertEqual(data.get("foo"), None)

        # setdefault
        self.assertEqual(data.setdefault("foo", 10), 10)
        self.assertEqual(data.setdefault("test", 10), 42)

        # update
        data.update({"foo": 5})
        self.assertEqual(data["foo"], 5)

        # pop
        self.assertEqual(data.pop("foo"), 5)
        self.assertEqual(data.model_dump(), {"test": 42})

        # popitem
        data["foo"] = 10
        self.assertEqual(data.popitem(), ("foo", 10))
        self.assertEqual(data.model_dump(), {"test": 42})

        # __delitem__
        del data["test"]
        self.assertEqual(data.model_dump(), {})

        # clear
        data["test"] = 42
        self.assertEqual(data.model_dump(), {"test": 42})
        data.clear()
        self.assertEqual(data.model_dump(), {})



from django.urls import include, path

from thelabtyping.api.routing import Router

from ..sampleapp.views import (
    router_users_detail,
    router_users_list,
    router_users_post,
)

# Alternate method of building a router
unnamespaced_router = Router(basename="unnamespaced-router")
unnamespaced_router.route(
    "users/",
    name="users-list",
    get=router_users_list,
    post=router_users_post,
)
unnamespaced_router.route(
    "users/<int:pk>/",
    name="users-detail",
    get=router_users_detail,
)

urlpatterns = [
    path("api/", include("thelabtyping.tests.sampleapp.urls")),
    path("unnamespaced-router/", include(unnamespaced_router.urls)),
]



from __future__ import annotations

from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from enum import StrEnum
from functools import reduce
from operator import or_
from typing import Concatenate, NewType, Protocol

from django.contrib.auth.decorators import permission_required
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model
from django.http import HttpRequest, HttpResponseBase, HttpResponseNotAllowed
from django.urls import URLPattern, path, reverse
from django.urls.exceptions import NoReverseMatch
import pydantic

from ..abc import DictOf
from .resource import CRUDViews
from .responses import APIResponse


class HttpMethod(StrEnum):
    """HTTP request methods."""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


URLPatternStr = NewType("URLPatternStr", str)
AllowedMethods = set[HttpMethod]

type ViewFn[**P, R: HttpResponseBase] = Callable[Concatenate[HttpRequest, P], R]
type ViewDecorator = Callable[
    [Callable[..., HttpResponseBase]],
    Callable[..., HttpResponseBase],
]

ALL_METHODS: AllowedMethods = {
    HttpMethod.GET,
    HttpMethod.POST,
    HttpMethod.PUT,
    HttpMethod.PATCH,
    HttpMethod.DELETE,
}


@dataclass
class RegisteredView[**P, R: HttpResponseBase]:
    """Container for a view function and its allowed HTTP methods."""

    methods: AllowedMethods
    fn: ViewFn[P, R]


@dataclass
class Route:
    """Represents a URL route with multiple HTTP method handlers."""

    name: str
    views: list[RegisteredView[[], HttpResponseBase]]
    decorators: Sequence[ViewDecorator]
    _router_decorators: Sequence[ViewDecorator]
    _cached_view: Callable[..., HttpResponseBase] | None

    def __init__(
        self,
        name: str,
        decorators: Sequence[ViewDecorator] | None = None,
        router_decorators: Sequence[ViewDecorator] | None = None,
    ) -> None:
        """Initialize a new route with the given name and optional decorators."""
        self.name = name
        self.views = []
        self.decorators = decorators or []
        self._router_decorators = router_decorators or []
        self._cached_view = None

    def get[**P, R: HttpResponseBase](self, fn: ViewFn[P, R]) -> ViewFn[P, R]:
        """Register a view function for GET requests."""
        return self.register({HttpMethod.GET})(fn)

    def post[**P, R: HttpResponseBase](self, fn: ViewFn[P, R]) -> ViewFn[P, R]:
        """Register a view function for POST requests."""
        return self.register({HttpMethod.POST})(fn)

    def put[**P, R: HttpResponseBase](self, fn: ViewFn[P, R]) -> ViewFn[P, R]:
        """Register a view function for PUT requests."""
        return self.register({HttpMethod.PUT})(fn)

    def patch[**P, R: HttpResponseBase](self, fn: ViewFn[P, R]) -> ViewFn[P, R]:
        """Register a view function for PATCH requests."""
        return self.register({HttpMethod.PATCH})(fn)

    def delete[**P, R: HttpResponseBase](self, fn: ViewFn[P, R]) -> ViewFn[P, R]:
        """Register a view function for DELETE requests."""
        return self.register({HttpMethod.DELETE})(fn)

    def register[**P, R: HttpResponseBase](
        self,
        methods: AllowedMethods = ALL_METHODS,
    ) -> Callable[
        [ViewFn[P, R]],
        ViewFn[P, R],
    ]:
        """Register a view function for the specified HTTP methods."""

        def decorator(view_fn: ViewFn[P, R]) -> ViewFn[P, R]:
            self.add_view(view_fn, methods)
            return view_fn

        return decorator

    def add_view[**P, R: HttpResponseBase](
        self,
        view_fn: ViewFn[P, R],
        methods: AllowedMethods,
    ) -> None:
        """Add a view function to this route for the given methods."""
        conflicts = self.allowed_methods & methods
        if conflicts:
            raise ImproperlyConfigured(
                f"Cannot {view_fn} for methods {conflicts}. Views already "
                "exist for these methods."
            )
        self.views.append(
            RegisteredView(
                methods=methods,
                fn=view_fn,
            )
        )

    def dispatch(
        self,
        request: HttpRequest,
        *args: object,
        **kwargs: object,
    ) -> HttpResponseBase:
        """Dispatch an incoming request to the appropriate view function."""
        for view in self.views:
            if request.method in view.methods:
                return view.fn(request, *args, **kwargs)

        return HttpResponseNotAllowed(self.allowed_methods)

    @property
    def allowed_methods(self) -> AllowedMethods:
        """Get all HTTP methods supported by this route."""
        return reduce(or_, (view.methods for view in self.views), set())

    @property
    def view(self) -> Callable[..., HttpResponseBase]:
        """Get dispatch with all decorators applied (router-level first, then route-level)."""
        if self._cached_view is None:
            # No decorators - return dispatch directly
            if not self.decorators and not self._router_decorators:
                self._cached_view = self.dispatch
            else:
                result: Callable[..., HttpResponseBase] = self.dispatch
                # Apply router-level decorators first (innermost)
                for decorator in self._router_decorators:
                    result = decorator(result)
                # Apply route-level decorators on top (outermost)
                for decorator in self.decorators:
                    result = decorator(result)
                self._cached_view = result
        return self._cached_view


type RouteMap = dict[URLPatternStr, Route]


class SchemaGenerator(Protocol):
    """Protocol for API schema generators (OpenAPI, etc.).

    Schema generators handle the full view logic including authentication,
    caching, and response serialization. The Router treats them as views.

    Example implementation:
        @dataclass
        class MySchemaGenerator:
            def __call__(
                self, router: Router, request: HttpRequest
            ) -> HttpResponseBase:
                spec = generate_my_schema(router.routes)
                return APIResponse(spec)
    """

    def __call__(self, router: Router, request: HttpRequest) -> HttpResponseBase:
        """Handle a schema request.

        Args:
            router: The Router instance (provides access to routes).
            request: The incoming HTTP request.

        Returns:
            An HTTP response (typically APIResponse with serialized schema).
        """
        ...


RouterIndex = DictOf[str, pydantic.HttpUrl]


class Router:
    """URL router for organizing and dispatching API routes."""

    basename: str | None = None
    routes: RouteMap
    decorators: Sequence[ViewDecorator]
    _schemas: Mapping[str, SchemaGenerator]

    def __init__(
        self,
        basename: str | None = None,
        enable_index: bool = True,
        schemas: Mapping[str, SchemaGenerator] | None = None,
        decorators: Sequence[ViewDecorator] | None = None,
    ) -> None:
        """Initialize a new router with optional basename, index view, and decorators."""
        self.basename = basename
        self.routes: RouteMap = {}
        self.decorators = decorators or []
        self._schemas = schemas or {}
        # Immediately register the root index view
        if enable_index:
            self.route("", name="index", get=self.index_view)

        # Register schema endpoints - generator handles ALL view logic (auth, caching, response)
        # Create a view that passes self (router) to the generator
        def make_schema_view(
            gen: SchemaGenerator,
        ) -> Callable[[HttpRequest], HttpResponseBase]:
            def view(request: HttpRequest) -> HttpResponseBase:
                return gen(self, request)

            return view

        for url_pattern, generator in self._schemas.items():
            route_name = url_pattern.lower().replace(".", "-")
            route_name = f"schema-{route_name}"
            self.route(
                url_pattern,
                name=route_name,
                get=make_schema_view(generator),
            )

    def index_view(self, request: HttpRequest) -> APIResponse[RouterIndex]:
        """Auto-generated index view listing all available routes."""
        index = RouterIndex({})
        namespace = request.resolver_match.namespace if request.resolver_match else None
        for pattern, route in self.routes.items():
            name = f"{namespace}:{route.name}" if namespace else route.name
            try:
                url_path = reverse(name)
            except NoReverseMatch:
                # Catch and ignore this so that we skip URLs which require
                # params (e.g. detail views)
                continue
            url = request.build_absolute_uri(url_path)
            index[name] = pydantic.HttpUrl(url)
        return APIResponse(index)

    def route[**P, R: HttpResponseBase](
        self,
        url_pattern: str,
        name: str,
        get: ViewFn[P, R] | None = None,
        post: ViewFn[P, R] | None = None,
        put: ViewFn[P, R] | None = None,
        patch: ViewFn[P, R] | None = None,
        delete: ViewFn[P, R] | None = None,
        decorators: Sequence[ViewDecorator] | None = None,
    ) -> Route:
        """Register a new route with the given URL pattern, view functions, and decorators."""
        _pattern = URLPatternStr(url_pattern)
        if _pattern in self.routes:
            raise ImproperlyConfigured(
                f"Cannot add route {_pattern} to router. Route already exists."
            )
        # Create route
        name = f"{self.basename}-{name}" if self.basename is not None else name
        route = Route(name, decorators=decorators, router_decorators=self.decorators)
        # Register any provided views
        views: dict[HttpMethod, ViewFn[P, R] | None] = {
            HttpMethod.GET: get,
            HttpMethod.POST: post,
            HttpMethod.PUT: put,
            HttpMethod.PATCH: patch,
            HttpMethod.DELETE: delete,
        }
        for method, view in views.items():
            if view is not None:
                route.register({method})(view)
        # Save and return the view
        self.routes[_pattern] = route
        return self.routes[_pattern]

    def register_resource[_ModelT: Model, _PKT](
        self,
        *,
        basename: str,
        list_pattern: str,
        detail_pattern: str,
        views: CRUDViews[_ModelT, _PKT],
        model: type[_ModelT] | None = None,
        list_permissions: Sequence[str] | None = None,
        detail_permissions: Sequence[str] | None = None,
        create_permissions: Sequence[str] | None = None,
        update_permissions: Sequence[str] | None = None,
        patch_permissions: Sequence[str] | None = None,
        delete_permissions: Sequence[str] | None = None,
        enable_list: bool = True,
        enable_detail: bool = True,
        enable_create: bool = True,
        enable_update: bool = True,
        enable_patch: bool = True,
        enable_delete: bool = True,
    ) -> tuple[Route | None, Route | None]:
        """Register CRUD views for a resource with automatic permission handling.

        Permission Resolution:
            When a permission parameter is None (not provided), permissions are
            inferred from Django's default model permissions using the model's
            app_label and model_name:
            - list_permissions: "{app_label}.view_{model_name}"
            - detail_permissions: "{app_label}.view_{model_name}"
            - create_permissions: "{app_label}.add_{model_name}"
            - update_permissions: "{app_label}.change_{model_name}"
            - patch_permissions: "{app_label}.change_{model_name}"
            - delete_permissions: "{app_label}.delete_{model_name}"

            To explicitly disable permission checks for an operation, pass an
            empty sequence (e.g., list_permissions=[]).

        Args:
            basename: Base name for URL route names (e.g., "user" -> "user-list")
            list_pattern: URL pattern for list/create endpoint
            detail_pattern: URL pattern for detail/update/delete endpoint
            views: CRUDViews NamedTuple from build_crud_resource_views
            model: Django model class for inferring default permissions. Required
                if any permission parameter is None.
            list_permissions: Permissions for GET list. None=infer, []=none
            detail_permissions: Permissions for GET detail. None=infer, []=none
            create_permissions: Permissions for POST. None=infer, []=none
            update_permissions: Permissions for PUT. None=infer, []=none
            patch_permissions: Permissions for PATCH. None=infer, []=none
            delete_permissions: Permissions for DELETE. None=infer, []=none
            enable_list: Whether to enable the list endpoint
            enable_detail: Whether to enable the detail endpoint
            enable_create: Whether to enable the create endpoint
            enable_update: Whether to enable the update endpoint
            enable_patch: Whether to enable the patch endpoint
            enable_delete: Whether to enable the delete endpoint

        Returns:
            Tuple of (list_route, detail_route), either may be None if disabled
        """

        def get_default_permission(action: str) -> list[str]:
            """Get default Django permission for an action (view/add/change/delete)."""
            if model is None:
                raise ImproperlyConfigured(
                    f"Cannot infer {action} permission: 'model' parameter is required "
                    "when permission parameters are None. Either provide the model "
                    "class or explicitly set permissions (use [] for no permissions)."
                )
            app_label = model._meta.app_label
            model_name = model._meta.model_name
            return [f"{app_label}.{action}_{model_name}"]

        def resolve_permissions(
            explicit: Sequence[str] | None,
            action: str,
        ) -> Sequence[str]:
            """Resolve permissions: explicit value, or infer from model."""
            if explicit is not None:
                return explicit
            return get_default_permission(action)

        def wrap_with_permissions(
            view: Callable[..., HttpResponseBase] | None,
            perms: Sequence[str],
        ) -> Callable[..., HttpResponseBase] | None:
            """Wrap a view with permission_required decorators."""
            if view is None:
                return None
            if not perms:
                return view
            result = view
            for perm in reversed(perms):
                result = permission_required(perm, raise_exception=True)(result)
            return result

        # Resolve all permissions (infer from model if None)
        resolved_list_perms = resolve_permissions(list_permissions, "view")
        resolved_detail_perms = resolve_permissions(detail_permissions, "view")
        resolved_create_perms = resolve_permissions(create_permissions, "add")
        resolved_update_perms = resolve_permissions(update_permissions, "change")
        resolved_patch_perms = resolve_permissions(patch_permissions, "change")
        resolved_delete_perms = resolve_permissions(delete_permissions, "delete")

        list_route: Route | None = None
        if enable_list or (enable_create and views.create_view):
            list_route = self.route(
                list_pattern,
                name=f"{basename}-list",
                get=(
                    wrap_with_permissions(views.list_view, resolved_list_perms)
                    if enable_list
                    else None
                ),
                post=(
                    wrap_with_permissions(views.create_view, resolved_create_perms)
                    if (enable_create and views.create_view)
                    else None
                ),
            )

        detail_route: Route | None = None
        if (
            enable_detail
            or (enable_update and views.update_view)
            or (enable_patch and views.patch_view)
            or enable_delete
        ):
            detail_route = self.route(
                detail_pattern,
                name=f"{basename}-detail",
                get=(
                    wrap_with_permissions(views.detail_view, resolved_detail_perms)
                    if enable_detail
                    else None
                ),
                put=(
                    wrap_with_permissions(views.update_view, resolved_update_perms)
                    if enable_update and views.update_view
                    else None
                ),
                patch=(
                    wrap_with_permissions(views.patch_view, resolved_patch_perms)
                    if enable_patch and views.patch_view
                    else None
                ),
                delete=(
                    wrap_with_permissions(views.delete_view, resolved_delete_perms)
                    if enable_delete
                    else None
                ),
            )

        return (list_route, detail_route)

    @property
    def urls(self) -> list[URLPattern]:
        """Get Django URLPattern objects for all registered routes."""
        return [
            path(pattern, route.view, name=route.name)
            for pattern, route in self.routes.items()
        ]



import json

from django.contrib.auth.models import User
from django.test import Client, TestCase
from django.urls import reverse

from thelabtyping.api.status import Status


class ClassBasedViewTest(TestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(username="joe", password="password")
        self.client = Client()

    def test_list_users(self) -> None:
        resp = self.client.get(reverse("sampleapp:users-list"))
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            [
                {
                    "username": "joe",
                    "first_name": "",
                    "last_name": "",
                }
            ],
        )

    def test_list_users_with_filter(self) -> None:
        resp = self.client.get(reverse("sampleapp:users-list") + "?first_name=Jim")
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(resp.content, [])

    def test_list_users_with_invalid_filter(self) -> None:
        resp = self.client.get(reverse("sampleapp:users-list") + "?id=FOO")
        self.assertEqual(resp.status_code, Status.HTTP_400_BAD_REQUEST)
        self.assertJSONEqual(
            resp.content,
            {
                "errors": {
                    "id": {
                        "msg": "Input should be a valid integer, unable to parse string as an integer",
                        "type": "int_parsing",
                    }
                }
            },
        )

    def test_create_user(self) -> None:
        self.client.login(username="joe", password="password")
        resp = self.client.post(
            reverse("sampleapp:users-list"),
            content_type="application/json",
            data={
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            {
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )

    def test_create_user_urlencoded(self) -> None:
        self.client.login(username="joe", password="password")
        resp = self.client.post(
            reverse("sampleapp:users-list"),
            data={
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )
        self.assertEqual(resp.status_code, Status.HTTP_200_OK)
        self.assertJSONEqual(
            resp.content,
            {
                "username": "jack",
                "first_name": "Jack",
                "last_name": "Jackson",
            },
        )

    def test_create_user_invalid_body(self) -> None:
        self.client.login(username="joe", password="password")
        resp = self.client.post(
            reverse("sampleapp:users-list"),
            content_type="application/json",
            data={
                "username": "jack",
                "first_name": "Jack",
            },
        )
        self.assertEqual(resp.status_code, Status.HTTP_400_BAD_REQUEST)
        self.assertJSONEqual(
            resp.content,
            {
                "errors": {
                    "last_name": {
                        "msg": "Field required",
                        "type": "missing",
                    },
                },
            },
        )

    def test_create_user_without_login(self) -> None:
        resp = self.client.post(
            reverse("sampleapp:users-list"),
            content_type="application/json",
            data=json.dumps(
                {
                    "username": "jack",
                    "first_name": "Jack",
                    "last_name": "Jackson",
                }
            ),
        )
        self.assertEqual(resp.status_code, Status.HTTP_403_FORBIDDEN)



"""End-to-end tests for OpenAPI generation with sample app."""

from typing import Any, ClassVar
from unittest import TestCase
import json
import urllib.request

from django.test import TestCase as DjangoTestCase
import jsonschema

from thelabtyping.abc import ListOf
from thelabtyping.api.openapi.generator import (
    OpenAPIGenerator,
    OpenAPIGeneratorConfig,
    generate_openapi,
)
from thelabtyping.api.requests import (
    AuthdTypedRequestBody,
    EmptyTypedRequest,
    TypedRequestQuery,
)
from thelabtyping.api.responses import APIResponse
from thelabtyping.api.routing import Router
from thelabtyping.api.serializers import Empty
from thelabtyping.api.views import validate
from thelabtyping.tests.sampleapp.views import router

from .conftest import BodyModel, OpenAPITestMixin, QueryModel, ResponseModel


class TestSampleAppOpenAPIGeneration(TestCase):
    def test_generate_spec_from_sample_router(self) -> None:
        """Generate OpenAPI spec from the sample app router."""
        spec = generate_openapi(
            router,
            title="Sample App API",
            version="1.0.0",
            description="Test API from sample app",
            namespace="sampleapp",
        )

        # Verify basic structure
        self.assertEqual(spec.openapi, "3.1.0")
        self.assertEqual(spec.info.title, "Sample App API")
        self.assertEqual(spec.info.version, "1.0.0")

    def test_sample_app_paths_exist(self) -> None:
        """Sample app routes are present in the generated spec."""
        spec = generate_openapi(
            router,
            title="Test",
            version="1.0.0",
            namespace="sampleapp",
        )

        # Check for expected paths
        self.assertIn("/api/router/users/", spec.paths)
        self.assertIn("/api/router/users/{pk}/", spec.paths)

    def test_users_list_endpoint(self) -> None:
        """Users list endpoint has GET and POST operations."""
        spec = generate_openapi(
            router,
            title="Test",
            version="1.0.0",
            namespace="sampleapp",
        )
        path_item = spec.paths["/api/router/users/"]
        self.assertIsNotNone(path_item.get)
        self.assertIsNotNone(path_item.post)

        # GET should have query parameters
        get_op = path_item.get
        assert get_op is not None  # for mypy
        self.assertIsNotNone(get_op.parameters)

        # POST should have request body and security (requires auth)
        post_op = path_item.post
        assert post_op is not None  # for mypy
        self.assertIsNotNone(post_op.requestBody)

    def test_user_detail_endpoint(self) -> None:
        """User detail endpoint has GET operation with path parameter."""
        spec = generate_openapi(
            router,
            title="Test",
            version="1.0.0",
            namespace="sampleapp",
        )

        path_item = spec.paths["/api/router/users/{pk}/"]
        self.assertIsNotNone(path_item.get)

        # Should have pk path parameter
        get_op = path_item.get
        assert get_op is not None  # for mypy
        self.assertIsNotNone(get_op.parameters)
        assert get_op.parameters is not None  # for mypy
        path_params = [
            p for p in get_op.parameters if getattr(p, "param_in", None) == "path"
        ]
        self.assertEqual(len(path_params), 1)
        self.assertEqual(path_params[0].name, "pk")  # type: ignore[union-attr]

    def test_components_schemas_generated(self) -> None:
        """Component schemas are generated for Pydantic models."""
        spec = generate_openapi(
            router,
            title="Test",
            version="1.0.0",
            namespace="sampleapp",
        )

        self.assertIsNotNone(spec.components)
        assert spec.components is not None  # for mypy
        self.assertIsNotNone(spec.components.schemas)
        assert spec.components.schemas is not None  # for mypy

        # Should have schemas for the sample app models
        schema_names = list(spec.components.schemas.keys())
        self.assertTrue(len(schema_names) > 0)

    def test_spec_serializes_to_valid_json(self) -> None:
        """Generated spec serializes to valid JSON."""
        spec = generate_openapi(
            router,
            title="Sample App API",
            version="1.0.0",
            description="Test API",
            namespace="sampleapp",
        )

        json_str = spec.model_dump_json(by_alias=True, exclude_none=True, indent=2)

        # Parse and verify
        data = json.loads(json_str)
        self.assertEqual(data["openapi"], "3.1.0")
        self.assertIn("paths", data)
        self.assertIn("components", data)

    def test_index_endpoint_included(self) -> None:
        """Router index endpoint is included if enabled."""
        spec = generate_openapi(
            router,
            title="Test",
            version="1.0.0",
            namespace="sampleapp",
        )

        # The sample app router has enable_index=True by default
        self.assertIn("/api/router/", spec.paths)


class TestOpenAPISpecStructure(TestCase):
    def test_responses_have_descriptions(self) -> None:
        """All responses have descriptions (required by OpenAPI)."""
        spec = generate_openapi(
            router,
            title="Test",
            version="1.0.0",
            namespace="sampleapp",
        )

        for path, path_item in spec.paths.items():
            for method in ["get", "post", "put", "patch", "delete"]:
                operation = getattr(path_item, method, None)
                if operation is not None:
                    for status_code, response in operation.responses.items():
                        self.assertIsNotNone(
                            response.description,
                            f"Response {status_code} for {method.upper()} {path} "
                            "has no description",
                        )

    def test_path_parameters_are_required(self) -> None:
        """Path parameters are marked as required."""
        spec = generate_openapi(
            router,
            title="Test",
            version="1.0.0",
            namespace="sampleapp",
        )

        for path, path_item in spec.paths.items():
            if "{" not in path:
                continue

            for method in ["get", "post", "put", "patch", "delete"]:
                operation = getattr(path_item, method, None)
                if operation is not None and operation.parameters:
                    path_params = [
                        p
                        for p in operation.parameters
                        if getattr(p, "param_in", None) == "path"
                    ]
                    for param in path_params:
                        self.assertTrue(
                            param.required,
                            f"Path parameter {param.name} in {path} should be required",
                        )

    def test_operation_ids_are_unique(self) -> None:
        """All operation IDs are unique across the spec."""
        spec = generate_openapi(
            router,
            title="Test",
            version="1.0.0",
            namespace="sampleapp",
        )

        operation_ids: list[str] = []
        for path_item in spec.paths.values():
            for method in ["get", "post", "put", "patch", "delete"]:
                operation = getattr(path_item, method, None)
                if operation is not None and operation.operationId:
                    operation_ids.append(operation.operationId)

        # Check for uniqueness
        self.assertEqual(
            len(operation_ids),
            len(set(operation_ids)),
            f"Duplicate operation IDs found: {operation_ids}",
        )


class TestOpenAPISchemaValidation(OpenAPITestMixin, DjangoTestCase):
    """Validate generated OpenAPI specs against the official OpenAPI 3.1 JSON schema."""

    OPENAPI_31_SCHEMA_URL = (
        "https://schemas.sourcemeta.com/openapi/v3.1/schema/2025-09-15.json"
    )
    _schema: ClassVar[dict[str, Any] | None] = None

    @classmethod
    def setUpClass(cls) -> None:
        """Fetch OpenAPI 3.1 JSON schema once for all tests."""
        super().setUpClass()
        with urllib.request.urlopen(cls.OPENAPI_31_SCHEMA_URL, timeout=30) as response:
            cls._schema = json.loads(response.read().decode())

    def _validate_spec(self, spec_data: dict[str, Any]) -> None:
        """Validate spec data against the OpenAPI 3.1 JSON schema."""
        assert self._schema is not None
        jsonschema.validate(spec_data, self._schema)

    def _generate_and_validate(self, test_router: Router, **kwargs: Any) -> None:
        """Generate OpenAPI spec from router and validate it."""
        spec = generate_openapi(
            test_router,
            title=kwargs.get("title", "Test API"),
            version=kwargs.get("version", "1.0.0"),
            description=kwargs.get("description"),
            servers=kwargs.get("servers"),
            namespace="sampleapp",
        )
        spec_data = json.loads(spec.model_dump_json(by_alias=True, exclude_none=True))
        self._validate_spec(spec_data)

    def test_sample_app_spec_validates(self) -> None:
        """Generated spec from sample app validates against OpenAPI 3.1 schema."""
        self._generate_and_validate(router)

    def test_minimal_spec_validates(self) -> None:
        """Minimal spec with empty router validates."""
        empty_router = Router(enable_index=False)
        self._generate_and_validate(empty_router)

    def test_spec_with_path_parameters_validates(self) -> None:
        """Spec with path parameters validates."""
        test_router = Router(enable_index=False)

        @test_router.route("users/<int:pk>/", name="user-detail").get
        @validate()
        def get_user(request: EmptyTypedRequest, pk: int) -> APIResponse[ResponseModel]:
            return APIResponse(ResponseModel(id=pk, name="test"))

        assert self._schema is not None
        self.validate_spec_for_test(test_router, self._schema)

    def test_spec_with_query_parameters_validates(self) -> None:
        """Spec with query parameters validates."""
        test_router = Router(enable_index=False)

        @test_router.route("search/", name="search").get
        @validate()
        def search(
            request: TypedRequestQuery[QueryModel],
        ) -> APIResponse[ListOf[ResponseModel]]:
            return APIResponse(ListOf[ResponseModel]([]))

        assert self._schema is not None
        self.validate_spec_for_test(test_router, self._schema)

    def test_spec_with_request_body_validates(self) -> None:
        """Spec with request body validates."""
        test_router = Router(enable_index=False)

        @test_router.route("users/", name="create-user").post
        @validate()
        def create_user(
            request: AuthdTypedRequestBody[BodyModel],
        ) -> APIResponse[ResponseModel]:
            return APIResponse(ResponseModel(id=1, name="test"))

        assert self._schema is not None
        self.validate_spec_for_test(test_router, self._schema)

    def test_spec_with_security_scheme_validates(self) -> None:
        """Spec with security scheme validates."""
        test_router = Router(enable_index=False)

        @test_router.route("protected/", name="protected").get
        @validate()
        def protected_view(
            request: AuthdTypedRequestBody[BodyModel],
        ) -> APIResponse[Empty]:
            return APIResponse(Empty())

        config = OpenAPIGeneratorConfig(
            title="Test API",
            version="1.0.0",
            auth_security_scheme="bearerAuth",
        )
        spec = self.generate_openapi_with_config_for_test(test_router, config)
        spec_data = json.loads(spec.model_dump_json(by_alias=True, exclude_none=True))
        self._validate_spec(spec_data)

    def test_spec_with_all_features_validates(self) -> None:
        """Spec with all features combined validates."""
        test_router = Router(enable_index=False)
        route = test_router.route("items/<int:item_id>/", name="item")

        @route.get
        @validate()
        def get_item(
            request: TypedRequestQuery[QueryModel], item_id: int
        ) -> APIResponse[ResponseModel]:
            return APIResponse(ResponseModel(id=item_id, name="test"))

        @route.put
        @validate()
        def update_item(
            request: AuthdTypedRequestBody[BodyModel], item_id: int
        ) -> APIResponse[ResponseModel]:
            return APIResponse(ResponseModel(id=item_id, name="updated"))

        @route.delete
        @validate()
        def delete_item(request: EmptyTypedRequest, item_id: int) -> APIResponse[Empty]:
            return APIResponse(Empty())

        config = OpenAPIGeneratorConfig(
            title="Full Feature API",
            version="2.0.0",
            description="API with all features",
            auth_security_scheme="bearerAuth",
            include_validation_errors=True,
        )
        spec = self.generate_openapi_with_config_for_test(test_router, config)
        spec_data = json.loads(spec.model_dump_json(by_alias=True, exclude_none=True))
        self._validate_spec(spec_data)


class TestOpenAPIPathResolutionWithNamespace(TestCase):
    """Tests that verify full path resolution with Django URL namespaces."""

    def test_generator_with_namespace_uses_full_paths(self) -> None:
        """Generator with namespace resolves full paths including base URL."""
        # This test verifies that when a namespace is provided,
        # the generator uses reverse() to get full paths
        config = OpenAPIGeneratorConfig(
            title="Test API",
            version="1.0.0",
        )
        # Use the sample app's namespace to test full path resolution
        generator = OpenAPIGenerator(config)
        spec = generator.generate(router, namespace="sampleapp")

        # Paths should include the full base path (/api/router/...)
        # when reverse() succeeds
        path_keys = list(spec.paths.keys())
        self.assertTrue(len(path_keys) > 0)

        # At least one path should have been resolved with the base path
        # The sample app router is included at /api/router/
        has_full_path = any(path.startswith("/api/router/") for path in path_keys)
        self.assertTrue(
            has_full_path,
            f"Expected paths to include base path /api/router/, got: {path_keys}",
        )

    def test_spec_paths_have_correct_parameter_syntax(self) -> None:
        """Paths in spec use OpenAPI parameter syntax {param}."""
        config = OpenAPIGeneratorConfig(
            title="Test API",
            version="1.0.0",
        )
        generator = OpenAPIGenerator(config)
        spec = generator.generate(router, namespace="sampleapp")

        # Find paths with parameters
        param_paths = [p for p in spec.paths.keys() if "{" in p]
        self.assertTrue(len(param_paths) > 0)

        # All parameter placeholders should use OpenAPI syntax
        for path in param_paths:
            # Should have {pk} format, not Django's <int:pk> format
            self.assertNotIn("<", path)
            self.assertNotIn(">", path)
            self.assertIn("{", path)
            self.assertIn("}", path)



from __future__ import annotations

from typing import Annotated
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from django.db.models import Model, QuerySet
from django.http import HttpRequest
import pydantic

# Maximum allowed page size to prevent excessive memory usage
MAX_PAGE_SIZE = 1000


class PaginationParams(pydantic.BaseModel):
    """Default pagination query parameters.

    Validation:
        - page: Must be >= 1 (1-indexed pagination)
        - page_size: Must be >= 1 and <= MAX_PAGE_SIZE (1000) when provided
    """

    page: Annotated[int, pydantic.Field(ge=1)] = 1
    page_size: Annotated[int | None, pydantic.Field(ge=1, le=MAX_PAGE_SIZE)] = None


class PaginatedResponse[T](pydantic.BaseModel):
    """Paginated API response with count, next/previous URLs, and results."""

    count: int
    next: str | None
    previous: str | None
    results: list[T]


def paginate_queryset[_ModelT: Model](
    request: HttpRequest,
    queryset: QuerySet[_ModelT],
    page: int,
    page_size: int,
) -> tuple[QuerySet[_ModelT], int, str | None, str | None]:
    """
    Paginate a queryset and return pagination metadata.

    Args:
        request: The HTTP request (used for building URLs)
        queryset: The queryset to paginate
        page: The page number (1-indexed)
        page_size: Number of items per page

    Returns:
        Tuple of (paginated_queryset, total_count, next_url, previous_url)
    """
    # Get total count
    count = queryset.count()

    # Calculate offset
    offset = (page - 1) * page_size

    # Slice queryset
    paginated_qs = queryset[offset : offset + page_size]

    # Calculate if there are more pages
    has_next = offset + page_size < count
    has_previous = page > 1

    # Build next/previous URLs
    next_url = _build_page_url(request, page + 1) if has_next else None
    previous_url = _build_page_url(request, page - 1) if has_previous else None

    return paginated_qs, count, next_url, previous_url


def _build_page_url(request: HttpRequest, page: int) -> str:
    """Build a URL for a specific page, preserving other query parameters."""
    # Parse the current URL
    parsed = urlparse(request.get_full_path())

    # Parse existing query parameters
    query_params = parse_qs(parsed.query)

    # Update the page parameter
    query_params["page"] = [str(page)]

    # Flatten the query params (parse_qs returns lists)
    flat_params = {k: v[0] if len(v) == 1 else v for k, v in query_params.items()}

    # Rebuild the URL
    new_query = urlencode(flat_params, doseq=True)
    new_parsed = parsed._replace(query=new_query)

    # Build absolute URL
    return request.build_absolute_uri(urlunparse(new_parsed))



from typing import TypedDict, assert_never
import json

from django.core.serializers.json import DjangoJSONEncoder
from django.http import HttpResponse
import pydantic
import pydantic_core

from ..result import Err, Ok
from .status import Status


class APIValidationError(TypedDict):
    """Structure for a single API validation error."""

    type: str
    msg: str


class APIValidationErrors(TypedDict):
    """Structure for API validation error response."""

    errors: dict[str, APIValidationError]


def _build_errors(exc: pydantic_core.ValidationError) -> APIValidationErrors:
    """Convert a Pydantic validation error to structured API error format."""
    errors = exc.errors(
        include_url=False,
        include_context=False,
        include_input=False,
    )
    resp: APIValidationErrors = {
        "errors": {
            (".".join(str(seg) for seg in err["loc"])): {
                "type": err["type"],
                "msg": err["msg"],
            }
            for err in errors
        }
    }
    return resp


class APIResponse[T: pydantic.BaseModel](HttpResponse):
    """
    An HTTP response class that accepts a Pydantic model object and sends it as
    JSON.
    """

    def __init__(
        self,
        # Accept either a model, or a Result which might contain a model.
        maybe_obj: T | Ok[T] | Err[pydantic_core.ValidationError] | Err[Exception],
        # Other response kwargs to pass through
        content_type: str = "application/json",
        status: int | None = None,
        reason: str | None = None,
        charset: str | None = None,
        headers: dict[str, str] | None = None,
        exclude_none: bool = False,
    ):
        """Create an API response from a Pydantic model or Result type."""
        data: str = ""
        if isinstance(maybe_obj, pydantic.BaseModel):
            obj = maybe_obj
            data = obj.model_dump_json(exclude_none=exclude_none)
        elif isinstance(maybe_obj, Ok):
            obj = maybe_obj.ok()
            data = obj.model_dump_json(exclude_none=exclude_none)
        elif isinstance(maybe_obj, Err):
            exc = maybe_obj.err()
            # If not a validation error, raise the error and let Django handle it
            if not isinstance(exc, pydantic_core.ValidationError):
                raise exc
            data = json.dumps(
                _build_errors(exc),
                indent=True,
                cls=DjangoJSONEncoder,
            )
            status = Status.HTTP_400_BAD_REQUEST
            reason = None
        else:
            assert_never(maybe_obj)  # pragma: no cover
        super().__init__(
            content=data,
            content_type=content_type,
            status=status,
            reason=reason,
            charset=charset,
            headers=headers,
        )



"""
Inspired in part by:

- https://github.com/rustedpy/result
- https://gcanti.github.io/fp-ts/

Provides an Either / Result type, as well as a decorator to convert normal
exception-throwing functions into functions which return a Result type.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Generator, Iterable, Iterator
from functools import wraps
from typing import Any, Literal, NoReturn, Self


class UnwrapError(TypeError):
    """Raised when attempting to unwrap a Result in the wrong state."""

    pass


class DoException[E](Exception):
    """
    This is used to signal to `do()` that the result is an `Err`,
    which short-circuits the generator and returns that Err.
    Using this exception for control flow in `do()` allows us
    to simulate `and_then()` in the Err case: namely, we don't call `op`,
    we just return `self` (the Err).
    """

    def __init__(self, err: Err[E]) -> None:
        self.err = err


class _Result[_OK, _Err](ABC):
    """Abstract base class for Result types."""

    ok_value: _OK
    err_value: _Err

    @property
    @abstractmethod
    def is_ok(self) -> bool: ...  # pragma: no cover

    @property
    @abstractmethod
    def is_err(self) -> bool: ...  # pragma: no cover

    @abstractmethod
    def ok(self) -> _OK | None: ...  # pragma: no cover

    @abstractmethod
    def err(self) -> _Err | None: ...  # pragma: no cover

    def unwrap(self) -> _OK:
        if self.is_ok:
            return self.ok_value
        raise UnwrapError("Called Either.unwrap() on an Err value")

    def unwrap_err(self) -> _Err:
        if self.is_err:
            return self.err_value
        raise UnwrapError("Called Either.unwrap_err() on an Ok value")


class Ok[_OK](_Result[_OK, None]):
    """Represents a successful result containing a value."""

    def __init__(self, value: _OK) -> None:
        self.ok_value = value
        self.err_value = None

    def __iter__(self) -> Iterator[_OK]:
        yield self.ok_value

    def __repr__(self) -> str:
        return f"Ok({repr(self.ok_value)})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Ok) and self.ok_value == other.ok_value

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    def __hash__(self) -> int:
        return hash((True, self.ok_value))

    @property
    def is_ok(self) -> Literal[True]:
        return True

    @property
    def is_err(self) -> Literal[False]:
        return False

    def ok(self) -> _OK:
        return self.ok_value

    def err(self) -> None:
        return None

    def and_then[L, R](self, op: Callable[[_OK], Result[L, R]]) -> Result[L, R]:
        return op(self.ok_value)


class Err[_Err](_Result[None, _Err]):
    """Represents a failed result containing an error."""

    def __init__(self, value: _Err) -> None:
        self.ok_value = None
        self.err_value = value

    def __iter__(self) -> Iterator[NoReturn]:
        def _iter() -> Iterator[NoReturn]:
            # Exception will be raised when the iterator is advanced, not
            # when it's created
            raise DoException(self)
            # This yield will never be reached, but is necessary to create a
            # generator
            yield

        return _iter()

    def __repr__(self) -> str:
        return f"Err({repr(self.err_value)})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Err) and self.err_value == other.err_value

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    def __hash__(self) -> int:
        return hash((False, self.err_value))

    @property
    def is_ok(self) -> Literal[False]:
        return False

    @property
    def is_err(self) -> Literal[True]:
        return True

    def ok(self) -> None:
        return None

    def err(self) -> _Err:
        return self.err_value

    def and_then(self, op: object) -> Self:
        return self


type Result[_OK, _Err] = Ok[_OK] | Err[_Err]


def as_result[**P, R, E: Exception](
    *catch: type[E],
) -> Callable[[Callable[P, R]], Callable[P, Result[R, E]]]:
    """Decorator that converts exception-throwing functions to return Result types."""

    def decorator(fn: Callable[P, R]) -> Callable[P, Result[R, E]]:
        @wraps(fn)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Result[R, E]:
            try:
                val = fn(*args, **kwargs)
                return Ok(val)
            except catch as e:
                return Err(e)

        return wrapper

    return decorator


def do[T, E](gen: Generator[Result[T, E]]) -> Result[T, E]:
    """
    Do notation for Result (syntactic sugar for sequence of `and_then()` calls).
    """
    try:
        return next(gen)
    except DoException as e:
        out: Err[E] = e.err
        return out


def partition_results[T, E](
    results: Iterable[Result[T, E]],
) -> tuple[
    list[T],
    list[E],
]:
    """
    Take an iterable of Results and partition them into two lists, one with all
    the OK values, one with all the Err values.
    """
    oks: list[T] = []
    errs: list[E] = []
    for res in results:
        if res.is_ok:
            oks.append(res.ok_value)
        else:
            errs.append(res.err_value)
    return oks, errs
